from typing import Union, Dict, List, Tuple, Callable, Any
from copy import deepcopy
from torch.distributions import Categorical
import torch as t
import torch.nn as nn
import numpy as np

from machin.frame.buffers.buffer import Transition, Buffer
from machin.frame.noise.action_space_noise import (
    add_normal_noise_to_action,
    add_clipped_normal_noise_to_action,
    add_uniform_noise_to_action,
    add_ou_noise_to_action,
)
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework, Config
from .utils import (
    hard_update,
    soft_update,
    safe_call,
    safe_return,
    assert_output_is_probs,
    assert_and_get_valid_models,
    assert_and_get_valid_optimizer,
    assert_and_get_valid_criterion,
    assert_and_get_valid_lr_scheduler,
)


class DDPG(TorchFramework):
    """
    DDPG framework.
    """

    _is_top = ["actor", "critic", "actor_target", "critic_target"]
    _is_restorable = ["actor_target", "critic_target"]

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        actor_target: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        critic_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple, Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict, Dict] = None,
        batch_size: int = 100,
        update_rate: float = 0.001,
        update_steps: Union[int, None] = None,
        actor_learning_rate: float = 0.0005,
        critic_learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        """
        Note:
            Your optimizer will be called as::

                optimizer(network.parameters(), learning_rate)

            Your lr_scheduler will be called as::

                lr_scheduler(
                    optimizer,
                    *lr_scheduler_args[0],
                    **lr_scheduler_kwargs[0],
                )

            Your criterion will be called as::

                criterion(
                    target_value.view(batch_size, 1),
                    predicted_value.view(batch_size, 1)
                )

        Note:
            DDPG supports two ways of updating the target network, the first
            way is polyak update (soft update), which updates the target network
            in every training step by mixing its weights with the online network
            using ``update_rate``.

            The other way is hard update, which copies weights of the online
            network after every ``update_steps`` training step.

            You can either specify ``update_rate`` or ``update_steps`` to select
            one update scheme, if both are specified, an error will be raised.

            These two different update schemes may result in different training
            stability.

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            update_steps: Training step number used to update target networks.
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
        """
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.update_steps = update_steps
        self.discount = discount
        self.gradient_max = gradient_max
        self.visualize = visualize
        self.visualize_dir = visualize_dir
        self._update_counter = 0

        if update_rate is not None and update_steps is not None:
            raise ValueError(
                "You can only specify one target network update"
                " scheme, either by update_rate or update_steps,"
                " but not both."
            )

        self.actor = actor
        self.actor_target = actor_target
        self.critic = critic
        self.critic_target = critic_target
        self.actor_optim = optimizer(self.actor.parameters(), lr=actor_learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(), lr=critic_learning_rate)
        self.replay_buffer = (
            Buffer(replay_size, replay_device)
            if replay_buffer is None
            else replay_buffer
        )

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.actor, self.actor_target)
            hard_update(self.critic, self.critic_target)

        if lr_scheduler is not None:
            if lr_scheduler_args is None:
                lr_scheduler_args = ((), ())
            if lr_scheduler_kwargs is None:
                lr_scheduler_kwargs = ({}, {})
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim, *lr_scheduler_args[0], **lr_scheduler_kwargs[0]
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim, *lr_scheduler_args[1], **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion
        super().__init__()

    @property
    def optimizers(self):
        return [self.actor_optim, self.critic_optim]

    @optimizers.setter
    def optimizers(self, optimizers):
        self.actor_optim, self.critic_optim = optimizers

    @property
    def lr_schedulers(self):
        if hasattr(self, "actor_lr_sch") and hasattr(self, "critic_lr_sch"):
            return [self.actor_lr_sch, self.critic_lr_sch]
        return []

    def act(self, state: Dict[str, Any], use_target: bool = False, **__):
        """
        Use actor network to produce an action for the current state.

        Args:
            state: Current state.
            use_target: Whether use the target network.

        Returns:
            Any thing returned by your actor network.
        """
        if use_target:
            return safe_return(safe_call(self.actor_target, state))
        else:
            return safe_return(safe_call(self.actor, state))

    def act_with_noise(
        self,
        state: Dict[str, Any],
        noise_param: Any = (0.0, 1.0),
        ratio: float = 1.0,
        mode: str = "uniform",
        use_target: bool = False,
        **__
    ):
        """
        Use actor network to produce a noisy action for the current state.

        See Also:
             :mod:`machin.frame.noise.action_space_noise`

        Args:
            state: Current state.
            noise_param: Noise params.
            ratio: Noise ratio.
            mode: Noise mode. Supported are:
                ``"uniform", "normal", "clipped_normal", "ou"``
            use_target: Whether use the target network.

        Returns:
            Noisy action of shape ``[batch_size, action_dim]``.
            Any other things returned by your actor network. if they exist.
        """
        if use_target:
            action, *others = safe_call(self.actor_target, state)
        else:
            action, *others = safe_call(self.actor, state)
        if mode == "uniform":
            noisy_action = add_uniform_noise_to_action(action, noise_param, ratio)
        elif mode == "normal":
            noisy_action = add_normal_noise_to_action(action, noise_param, ratio)
        elif mode == "clipped_normal":
            noisy_action = add_clipped_normal_noise_to_action(
                action, noise_param, ratio
            )
        elif mode == "ou":
            noisy_action = add_ou_noise_to_action(action, noise_param, ratio)
        else:
            raise ValueError("Unknown noise type: " + str(mode))

        if len(others) == 0:
            return noisy_action
        else:
            return (noisy_action, *others)

    def act_discrete(self, state: Dict[str, Any], use_target: bool = False, **__):
        """
        Use actor network to produce a discrete action for the current state.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Action of shape ``[batch_size, 1]``.
            Action probability tensor of shape ``[batch_size, action_num]``,
            produced by your actor.
            Any other things returned by your Q network. if they exist.
        """
        if use_target:
            action, *others = safe_call(self.actor_target, state)
        else:
            action, *others = safe_call(self.actor, state)

        assert_output_is_probs(action)
        batch_size = action.shape[0]
        result = t.argmax(action, dim=1).view(batch_size, 1)
        return (result, action, *others)

    def act_discrete_with_noise(
        self,
        state: Dict[str, Any],
        use_target: bool = False,
        choose_max_prob: float = 0.95,
        **__
    ):
        """
        Use actor network to produce a noisy discrete action for
        the current state.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            state: Current state.
            use_target: Whether to use the target network.
            choose_max_prob: Probability to choose the largest component when actor
                is outputing extreme probability vector like ``[0, 1, 0, 0]``.

        Returns:
            Noisy action of shape ``[batch_size, 1]``.
            Action probability tensor of shape ``[batch_size, action_num]``.
            Any other things returned by your Q network. if they exist.
        """
        if use_target:
            action, *others = safe_call(self.actor_target, state)
        else:
            action, *others = safe_call(self.actor, state)

        assert_output_is_probs(action)
        batch_size = action.shape[0]
        action_dim = action.shape[1]
        if action_dim > 1 and choose_max_prob < 1.0:
            scale = np.log((action_dim - 1) / (1 - choose_max_prob) * choose_max_prob)
            action = t.softmax(action * scale, dim=1)
        dist = Categorical(action)

        result = dist.sample([batch_size, 1]).view(batch_size, 1)
        return (result, action, *others)

    def _act(self, state: Dict[str, Any], use_target: bool = False, **__):
        """
        Use actor network to produce an action for the current state.

        Args:
            state: Current state.
            use_target: Whether use the target network.

        Returns:
            Action of shape ``[batch_size, action_dim]``.
        """
        if use_target:
            return safe_call(self.actor_target, state)[0]
        else:
            return safe_call(self.actor, state)[0]

    def _criticize(
        self,
        state: Dict[str, Any],
        action: Dict[str, Any],
        use_target: bool = False,
        **__
    ):
        """
        Use critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Q Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(self.critic_target, state, action)[0]
        else:
            return safe_call(self.critic, state, action)[0]

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(
            transition,
            required_attrs=("state", "action", "reward", "next_state", "terminal"),
        )

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(
                trans,
                required_attrs=("state", "action", "reward", "next_state", "terminal"),
            )

    def update(
        self,
        update_value=True,
        update_policy=True,
        update_target=True,
        concatenate_samples=True,
        **__
    ):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.
            update_target: Whether to update targets.
            concatenate_samples: Whether to concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        self.actor.train()
        self.critic.train()
        (
            batch_size,
            (state, action, reward, next_state, terminal, others,),
        ) = self.replay_buffer.sample_batch(
            self.batch_size,
            concatenate_samples,
            sample_method="random_unique",
            sample_attrs=["state", "action", "reward", "next_state", "terminal", "*"],
        )

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_function(
                self._act(next_state, True), next_state, others
            )
            next_value = self._criticize(next_state, next_action, True)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_function(
                reward, self.discount, next_value, terminal, others
            )

        cur_value = self._criticize(state, action)
        value_loss = self.criterion(cur_value, y_i.type_as(cur_value))

        if self.visualize:
            self.visualize_model(value_loss, "critic", self.visualize_dir)

        if update_value:
            self.critic.zero_grad()
            self._backward(value_loss)
            nn.utils.clip_grad_norm_(self.critic.parameters(), self.gradient_max)
            self.critic_optim.step()

        # Update actor network
        cur_action = self.action_transform_function(self._act(state), state, others)
        act_value, *_ = self._criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor", self.visualize_dir)

        if update_policy:
            self.actor.zero_grad()
            self._backward(act_policy_loss)
            nn.utils.clip_grad_norm_(self.actor.parameters(), self.gradient_max)
            self.actor_optim.step()

        # Update target networks
        if update_target:
            if self.update_rate is not None:
                soft_update(self.actor_target, self.actor, self.update_rate)
                soft_update(self.critic_target, self.critic, self.update_rate)
            else:
                self._update_counter += 1
                if self._update_counter % self.update_steps == 0:
                    hard_update(self.actor_target, self.actor)
                    hard_update(self.critic_target, self.critic)

        self.actor.eval()
        self.critic.eval()
        # use .item() to prevent memory leakage
        return -act_policy_loss.item(), value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()

    def load(
        self, model_dir: str, network_map: Dict[str, str] = None, version: int = -1
    ):
        # DOC INHERITED
        super().load(model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.actor, self.actor_target)
            hard_update(self.critic, self.critic_target)

    @staticmethod
    def action_transform_function(raw_output_action: Any, *_):
        """
        The action transform function is used to transform the output
        of actor to the input of critic.
        Action transform function must accept:

          1. Raw action from the actor model.
          2. Concatenated :attr:`.Transition.next_state`.
          3. Any other concatenated lists of custom keys from \
              :class:`.Transition`.

        and returns:
          1. A dictionary with the same form as :attr:`.Transition.action`

        Args:
          raw_output_action: Raw action from the actor model.
        """
        return {"action": raw_output_action}

    @staticmethod
    def reward_function(reward, discount, next_value, terminal, _):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * ~terminal * next_value

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "models": ["Actor", "Actor", "Critic", "Critic"],
            "model_args": ((), (), (), ()),
            "model_kwargs": ({}, {}, {}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "update_rate": 0.001,
            "update_steps": None,
            "actor_learning_rate": 0.0005,
            "critic_learning_rate": 0.001,
            "gradient_max": np.inf,
            "discount": 0.99,
            "replay_size": 500000,
            "replay_device": "cpu",
            "replay_buffer": None,
            "visualize": False,
            "visualize_dir": "",
        }
        config = deepcopy(config)
        config["frame"] = "DDPG"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = config["frame_config"]
        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*models, **f_config)
        return frame



from copy import deepcopy
from typing import Dict, Any, Union
from machin.frame.algorithms import TorchFramework
from machin.utils.conf import Config
from . import envs
import inspect
import torch as t
import machin.frame.algorithms as algorithms


def fill_default(
    default: Union[Dict[str, Any], Config], config: Union[Dict[str, Any], Config]
):
    for key in default:
        if key not in config:
            config[key] = default[key]
    return config


def get_available_algorithms():
    algos = []
    for algo in dir(algorithms):
        algo_cls = getattr(algorithms, algo)
        if (
            inspect.isclass(algo_cls)
            and issubclass(algo_cls, TorchFramework)
            and algo_cls != TorchFramework
        ):
            algos.append(algo)
    return algos


def get_available_environments():
    environments = []
    for e in dir(envs):
        e_module = getattr(envs, e)
        if hasattr(e_module, "launch") and hasattr(e_module, "generate_env_config"):
            environments.append(e)
    return environments


def generate_training_config(
    root_dir: str = "trial",
    episode_per_epoch: int = 10,
    max_episodes: int = 10000,
    config: Union[Dict[str, Any], Config] = None,
):
    config = deepcopy(config) or {}
    config["root_dir"] = root_dir
    config["episode_per_epoch"] = episode_per_epoch
    config["max_episodes"] = max_episodes
    config["early_stopping_patience"] = 3
    return config


def generate_algorithm_config(
    algorithm: str, config: Union[Dict[str, Any], Config] = None
):
    config = deepcopy(config) or {}
    if hasattr(algorithms, algorithm):
        algo_obj = getattr(algorithms, algorithm)
        if inspect.isclass(algo_obj) and issubclass(algo_obj, TorchFramework):
            config = algo_obj.generate_config(config)
            if algo_obj.is_distributed():
                # in pytorch lightning, gpus will override num_processes
                config["gpus"] = [0, 0, 0]
                config["num_processes"] = 3
                config["num_nodes"] = 1
            else:
                config["gpus"] = [0]
            return config
    raise ValueError(
        f"Invalid algorithm: {algorithm}, valid ones are: {get_available_algorithms()}"
    )


def generate_env_config(environment: str, config: Union[Dict[str, Any], Config] = None):
    config = deepcopy(config) or {}
    if hasattr(envs, environment):
        e_module = getattr(envs, environment)
        if hasattr(e_module, "launch") and hasattr(e_module, "generate_env_config"):
            return e_module.generate_env_config(config)
    raise ValueError(
        f"Invalid environment: {environment}, "
        f"valid ones are: {get_available_algorithms()}"
    )


def init_algorithm_from_config(
    config: Union[Dict[str, Any], Config], model_device: Union[str, t.device] = "cpu"
):
    assert_algorithm_config_complete(config)
    frame = getattr(algorithms, config["frame"], None)
    if not inspect.isclass(frame) or not issubclass(frame, TorchFramework):
        raise ValueError(
            f"Invalid algorithm: {config['frame']}, "
            f"valid ones are: {get_available_algorithms()}"
        )
    return frame.init_from_config(config, model_device=model_device)


def is_algorithm_distributed(config: Union[Dict[str, Any], Config]):
    assert_algorithm_config_complete(config)
    frame = getattr(algorithms, config["frame"], None)
    if not inspect.isclass(frame) or not issubclass(frame, TorchFramework):
        raise ValueError(
            f"Invalid algorithm: {config['frame']}, "
            f"valid ones are: {get_available_algorithms()}"
        )
    return frame.is_distributed()


def assert_training_config_complete(config: Union[Dict[str, Any], Config]):
    assert "root_dir" in config, 'Missing key "root_dir"'
    assert "episode_per_epoch" in config, 'Missing key "episode_per_epoch"'
    assert "max_episodes" in config, 'Missing key "max_episodes"'
    assert "early_stopping_patience" in config, 'Missing key "early_stopping_patience"'


def assert_algorithm_config_complete(config: Union[Dict[str, Any], Config]):
    assert "frame" in config, 'Missing key "frame" in config.'
    assert "frame_config" in config, 'Missing key "frame_config" in config.'


def assert_env_config_complete(config: Union[Dict[str, Any], Config]):
    assert "env" in config, 'Missing key "env" ' "in config."
    assert "train_env_config" in config, 'Missing key "train_env_config" ' "in config."
    assert "test_env_config" in config, 'Missing key "test_env_config" ' "in config."


def launch(config: Union[Dict[str, Any], Config]):
    assert_training_config_complete(config)
    assert_env_config_complete(config)
    assert_algorithm_config_complete(config)
    e_module = getattr(envs, config["env"])
    return e_module.launch(config)



from .a2c import *
from machin.parallel.server import PushPullGradServer
from machin.frame.helpers.servers import grad_server_helper
from .utils import FakeOptimizer, assert_and_get_valid_lr_scheduler


class A3C(A2C):
    """
    A3C framework.
    """

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        criterion: Callable,
        grad_server: Tuple[PushPullGradServer, PushPullGradServer],
        *_,
        batch_size: int = 100,
        actor_update_times: int = 5,
        critic_update_times: int = 10,
        entropy_weight: float = None,
        value_weight: float = 0.5,
        gradient_max: float = np.inf,
        gae_lambda: float = 1.0,
        discount: float = 0.99,
        normalize_advantage: bool = True,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        """
        See Also:
            :class:`.A2C`

        Note:
            A3C algorithm relies on parameter servers to synchronize
            parameters of actor and critic models across samplers (
            interact with environment) and trainers (using samples
            to train.

            The parameter server type :class:`.PushPullGradServer`
            used here utilizes gradients calculated by trainers:

            1. perform a "sum" reduction process on the collected
            gradients, then apply this reduced gradient to the model
            managed by its primary reducer

            2. push the parameters of this updated managed model to
            a ordered key-value server so that all processes,
            including samplers and trainers, can access the updated
            parameters.

            The ``grad_servers`` argument is a pair of accessors to
            two :class:`.PushPullGradServerImpl` class.

        Args:
            actor: Actor network module.
            critic: Critic network module.
            criterion: Criterion used to evaluate the value loss.
            grad_server: Custom gradient sync server accessors, the first
                server accessor is for actor, and the second one is for critic.
            batch_size: Batch size used during training.
            actor_update_times: Times to update actor in ``update()``.
            critic_update_times: Times to update critic in ``update()``.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            normalize_advantage: Whether to normalize the advantage function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
        """
        # Adam is just a placeholder here, the actual optimizer is
        # set in parameter servers
        super().__init__(
            actor,
            critic,
            FakeOptimizer,
            criterion,
            batch_size=batch_size,
            actor_update_times=actor_update_times,
            critic_update_times=critic_update_times,
            entropy_weight=entropy_weight,
            value_weight=value_weight,
            gradient_max=gradient_max,
            gae_lambda=gae_lambda,
            discount=discount,
            normalize_advantage=normalize_advantage,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=replay_buffer,
            visualize=visualize,
            visualize_dir=visualize_dir,
        )
        # disable local stepping
        self.actor_optim.step = lambda: None
        self.critic_optim.step = lambda: None
        self.actor_grad_server, self.critic_grad_server = grad_server[0], grad_server[1]
        self.is_syncing = True

    @property
    def optimizers(self):
        return []

    @optimizers.setter
    def optimizers(self, optimizers):
        pass

    @property
    def lr_schedulers(self):
        return []

    @classmethod
    def is_distributed(cls):
        return True

    def set_sync(self, is_syncing):
        self.is_syncing = is_syncing

    def manual_sync(self):
        self.actor_grad_server.pull(self.actor)
        self.critic_grad_server.pull(self.critic)

    def act(self, state: Dict[str, Any], **__):
        # DOC INHERITED
        if self.is_syncing:
            self.actor_grad_server.pull(self.actor)
        return super().act(state)

    def _eval_act(self, state: Dict[str, Any], action: Dict[str, Any], **__):
        # DOC INHERITED
        if self.is_syncing:
            self.actor_grad_server.pull(self.actor)
        return super()._eval_act(state, action)

    def _criticize(self, state: Dict[str, Any], *_, **__):
        # DOC INHERITED
        if self.is_syncing:
            self.critic_grad_server.pull(self.critic)
        return super()._criticize(state)

    def update(
        self, update_value=True, update_policy=True, concatenate_samples=True, **__
    ):
        # DOC INHERITED
        org_sync = self.is_syncing
        self.is_syncing = False
        super().update(update_value, update_policy, concatenate_samples)
        self.is_syncing = org_sync
        self.actor_grad_server.push(self.actor)
        self.critic_grad_server.push(self.critic)

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "grad_server_group_name": "a3c_grad_server",
            "grad_server_members": "all",
            "models": ["Actor", "Critic"],
            "model_args": ((), ()),
            "model_kwargs": ({}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "actor_update_times": 5,
            "critic_update_times": 10,
            "actor_learning_rate": 0.001,
            "critic_learning_rate": 0.001,
            "entropy_weight": None,
            "value_weight": 0.5,
            "gradient_max": np.inf,
            "gae_lambda": 1.0,
            "discount": 0.99,
            "normalize_advantage": True,
            "replay_size": 500000,
            "replay_device": "cpu",
            "replay_buffer": None,
            "visualize": False,
            "visualize_dir": "",
        }
        config = deepcopy(config)
        config["frame"] = "A3C"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = deepcopy(config["frame_config"])
        model_cls = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(model_cls, model_args, model_kwargs)
        ]
        model_creators = [
            lambda: m(*arg, **kwarg)
            for m, arg, kwarg in zip(model_cls, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )

        servers = grad_server_helper(
            model_creators,
            group_name=f_config["grad_server_group_name"],
            members=f_config["grad_server_members"],
            optimizer=optimizer,
            learning_rate=[
                f_config["actor_learning_rate"],
                f_config["critic_learning_rate"],
            ],
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=f_config["lr_scheduler_args"] or ((), ()),
            lr_scheduler_kwargs=f_config["lr_scheduler_kwargs"] or ({}, {}),
        )
        del f_config["criterion"]
        frame = cls(*models, criterion, servers, **f_config)
        return frame



from datetime import datetime, timedelta
from typing import Union, Iterable
from os.path import join

import os
import shutil

from machin.utils.logging import default_logger
from machin.utils.prepare import prep_clear_dirs, prep_create_dirs


class SaveEnv:
    def __init__(
        self,
        env_root: str,
        restart_from_trial: Union[str, None] = None,
        time_format="%Y_%m_%d_%H_%M_%S",
        create_sub_dirs=True,
    ):
        """
        Create the default environment for saving. creates something like::

            <your environment root>
            ├── config
            ├── log
            │   ├── images
            │   └── train_log
            └── model

        Args:
            env_root: root directory for all trials of the environment.
            restart_from_trial: instead of creating a new save environment
                for a new trial, use a existing save environment of an older
                trial, old trial name should be in format ``time_format``
            time_format: Time formatter, setting it to an empty string will cause
                the save environment to use ``env_root`` directly instead of using
                sub directories with a datetime name.
            create_dirs: Whether to create directories.
        """
        self.env_root = env_root
        self.time_format = time_format

        if restart_from_trial is None:
            self.env_create_time = datetime.now()
        else:
            self.env_create_time = datetime.strptime(
                restart_from_trial, self.time_format
            )
        self._check_dirs()
        if create_sub_dirs:
            self.create_sub_dirs()

    def create_dirs(self, dirs: Iterable[str]):
        """
        Create additional directories in root.

        Args:
            dirs: Directories.
        """
        prep_create_dirs(
            [
                join(self.env_root, self.env_create_time.strftime(self.time_format), d)
                for d in dirs
            ]
        )

    def get_trial_root(self):
        # pylint: disable=missing-docstring
        return join(self.env_root, self.env_create_time.strftime(self.time_format))

    def get_trial_config_dir(self):
        # pylint: disable=missing-docstring
        return join(
            self.env_root, self.env_create_time.strftime(self.time_format), "config"
        )

    def get_trial_model_dir(self):
        # pylint: disable=missing-docstring
        return join(
            self.env_root, self.env_create_time.strftime(self.time_format), "model"
        )

    def get_trial_image_dir(self):
        # pylint: disable=missing-docstring
        return join(
            self.env_root,
            self.env_create_time.strftime(self.time_format),
            "log",
            "images",
        )

    def get_trial_train_log_dir(self):
        # pylint: disable=missing-docstring
        return join(
            self.env_root,
            self.env_create_time.strftime(self.time_format),
            "log",
            "train_log",
        )

    def get_trial_time(self):
        # pylint: disable=missing-docstring
        return self.env_create_time

    def clear_trial_config_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs(
            [
                join(
                    self.env_root,
                    self.env_create_time.strftime(self.time_format),
                    "config",
                )
            ]
        )

    def clear_trial_model_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs(
            [
                join(
                    self.env_root,
                    self.env_create_time.strftime(self.time_format),
                    "model",
                )
            ]
        )

    def clear_trial_image_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs(
            [
                join(
                    self.env_root,
                    self.env_create_time.strftime(self.time_format),
                    "log",
                    "images",
                )
            ]
        )

    def clear_trial_train_log_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs(
            [
                join(
                    self.env_root,
                    self.env_create_time.strftime(self.time_format),
                    "log",
                    "train_log",
                )
            ]
        )

    def remove_trials_older_than(
        self,
        diff_day: int = 0,
        diff_hour: int = 1,
        diff_minute: int = 0,
        diff_second: int = 0,
    ):
        """
        By default this function removes all trials started one hour earlier
        than current time.

        Args:
            diff_day: Difference in days.
            diff_hour: Difference in hours.
            diff_minute: Difference in minutes.
            diff_second: Difference in seconds.
        """
        trial_list = [f for f in os.listdir(self.env_root)]
        current_time = datetime.now()
        diff_threshold = timedelta(
            days=diff_day, hours=diff_hour, minutes=diff_minute, seconds=diff_second
        )
        for file in trial_list:
            try:
                time = datetime.strptime(file, self.time_format)
            except ValueError:
                # not a trial
                pass
            else:
                diff_time = current_time - time
                if diff_time > diff_threshold:
                    rm_path = join(self.env_root, file)
                    default_logger.info(f"Removing trial directory: {rm_path}")
                    shutil.rmtree(rm_path)

    def create_sub_dirs(self):
        root_dir = join(self.env_root, self.env_create_time.strftime(self.time_format))
        prep_create_dirs(
            (
                join(root_dir, "model"),
                join(root_dir, "config"),
                join(root_dir, "log", "images"),
                join(root_dir, "log", "train_log"),
            )
        )

    def _check_dirs(self):
        """
        Overload this function in your environment class to check directory
        mapping

        Raises:
            RuntimeError if directory mapping is invalid.
        """
        pass



import os
import logging
import pytorch_lightning as pl
from time import sleep
from torch import distributed
from pytorch_lightning.utilities.seed import seed_everything
from pytorch_lightning.utilities.distributed import rank_zero_only
from pytorch_lightning.plugins.training_type.ddp import DDPPlugin as DDP
from pytorch_lightning.plugins.training_type.ddp_spawn import DDPSpawnPlugin as DDPS
from machin.parallel.distributed import World, is_world_initialized
from machin.parallel.process import Process
import multiprocessing as mp
import traceback

pl_logger = logging.getLogger("lightning")


def assert_world_config(cluster_env, master_addr, master_port, world_size):
    assert master_addr is not None, (
        "Master address not set, please set environment "
        "variable MASTER_ADDR or refer to the cluster environment "
        f"instructions you are using, current cluster env is {cluster_env}."
    )
    assert master_port is not None, (
        "Master port not set, please set environment "
        "variable MASTER_PORT or refer to the cluster environment "
        f"instructions you are using, current cluster env is {cluster_env}."
    )
    assert world_size is not None, (
        "World size not set, please set environment "
        "variable WORLD_SIZE or refer to the cluster environment "
        f"instructions you are using, current cluster env is {cluster_env}."
    )


class DDPPlugin(DDP):
    def init_ddp_connection(self, global_rank: int, world_size: int) -> None:
        master_addr = self.cluster_environment.master_address()
        master_port = self.cluster_environment.master_port()
        world_size = self.cluster_environment.world_size()
        assert_world_config(
            self.cluster_environment, master_addr, master_port, world_size
        )

        if not is_world_initialized():
            pl_logger.info(
                f"initializing world: GLOBAL_RANK: {global_rank}, "
                f"MEMBER: {int(global_rank) + 1}/{world_size}"
            )
            # TODO: currently nccl is having problems with supporting
            # different cnfigurations, use gloo as replacement.
            # See: https://github.com/pytorch/pytorch/issues/47885
            _w = World(
                name=str(global_rank),
                rank=int(global_rank),
                world_size=int(world_size),
                dist_backend="gloo",
                dist_init_method=f"tcp://{master_addr}:{master_port}",
                rpc_init_method=f"tcp://{master_addr}:{int(master_port) + 1}",
            )

    def pre_dispatch(self):
        """
        This function is called by the trainer, before dispatch(), which
        starts training/testing/... etc.

        DDP plugin will initialize ddp connection at here.
        """

        # TODO: check if needed
        seed = os.environ.get("PL_GLOBAL_SEED")
        if seed is not None:
            seed_everything(int(seed))

        # determine which process we are and world size
        self.set_world_ranks()

        # set warning rank
        rank_zero_only.rank = self.global_rank

        # set up server using proc 0's ip address
        # try to init for 20 times at max in case ports are taken
        # where to store ip_table
        self.init_ddp_connection(self.global_rank, self.world_size)

        # on world_size=0 let everyone know training is starting
        if self.is_global_zero and distributed.is_initialized():
            pl_logger.info("-" * 100)
            pl_logger.info(f"distributed_backend={self.distributed_backend}")
            pl_logger.info(
                f"All DDP processes registered. Starting ddp with {self.world_size} "
                "processes"
            )
            pl_logger.info("-" * 100)

        # set the ranks and devices
        self.dist.rank = self.global_rank
        self.dist.device = self.root_device

        if self.sync_batchnorm:
            self.model = self.configure_sync_batchnorm(self.model)

        # Move the launcher to the correct device
        # So in launcher, self.device would be set.
        self.model_to_device()

        # initialize framework in the launcher
        self._model.init_frame()
        if self._model.frame.optimizers is not None:
            self._model.trainer.accelerator.optimizers = self._model.frame.optimizers
        if self._model.frame.lr_schedulers is not None:
            self._model.trainer.accelerator.lr_schedulers = (
                self._model.frame.lr_schedulers
            )

        self.barrier()

    def training_step(self, *args, **kwargs):
        return self.lightning_module.training_step(*args, **kwargs)

    def validation_step(self, *args, **kwargs):
        return self.lightning_module.validation_step(*args, **kwargs)

    def test_step(self, *args, **kwargs):
        return self.lightning_module.test_step(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.lightning_module.predict(*args, **kwargs)

    def post_training_step(self):
        pass


class DDPSpawnPlugin(DDPS):
    def init_ddp_connection(self, global_rank: int, world_size: int) -> None:
        master_addr = self.cluster_environment.master_address()
        master_port = self.cluster_environment.master_port()
        world_size = self.cluster_environment.world_size()
        assert_world_config(
            self.cluster_environment, master_addr, master_port, world_size
        )

        if not is_world_initialized():
            pl_logger.info(
                f"initializing world: GLOBAL_RANK: {global_rank}, "
                f"MEMBER: {int(global_rank) + 1}/{world_size}"
            )
            # TODO: currently nccl is having problems with supporting
            # different cnfigurations, use gloo as replacement.
            # See: https://github.com/pytorch/pytorch/issues/47885
            _w = World(
                name=str(global_rank),
                rank=int(global_rank),
                world_size=int(world_size),
                dist_backend="gloo",
                dist_init_method=f"tcp://{master_addr}:{master_port}",
                rpc_init_method=f"tcp://{master_addr}:{int(master_port) + 1}",
            )

    def new_process(self, process_idx, trainer, mp_queue):
        """
        DDPSpawn does not pre-create processes. Instead it spawns processes as needed
        when training, testing, etc. is started, therefore we need to overload the
        process main function.
        """
        global trace_enabled
        self.mp_queue = mp_queue

        # TODO: check if needed
        seed = os.environ.get("PL_GLOBAL_SEED")
        if seed is not None:
            seed_everything(int(seed))

        self.set_world_ranks(process_idx)

        # set warning rank
        rank_zero_only.rank = self.global_rank

        # set up server using proc 0's ip address
        # try to init for 20 times at max in case ports are taken
        # where to store ip_table
        self.init_ddp_connection(self.global_rank, self.world_size)

        # on world_size=0 let everyone know training is starting
        if self.is_global_zero and distributed.is_initialized():
            pl_logger.info("-" * 100)
            pl_logger.info(f"distributed_backend={self.distributed_backend}")
            pl_logger.info(
                f"All DDP processes registered. Starting ddp with {self.world_size} "
                "processes"
            )
            pl_logger.info("-" * 100)

        # set the ranks and devices
        self.dist.rank = self.global_rank
        self.dist.device = self.root_device

        if self.sync_batchnorm:
            self.model = self.configure_sync_batchnorm(self.model)

        # Move the launcher to the correct device
        # So in launcher, self.device would be set.
        self.model_to_device()

        # initialize framework in the launcher
        self._model.init_frame()
        if self._model.frame.optimizers is not None:
            self._model.trainer.accelerator.optimizers = self._model.frame.optimizers
        if self._model.frame.lr_schedulers is not None:
            self._model.trainer.accelerator.lr_schedulers = (
                self._model.frame.lr_schedulers
            )

        self.barrier()

        results = trainer.train_or_test_or_predict()

        # persist info in ddp_spawn
        self.transfer_distrib_spawn_state_on_fit_end(results)

    def start_training(self, trainer):
        self._spawn()
        trainer.optimizers = []

    def start_testing(self, trainer):
        self._spawn()

    def start_predicting(self, trainer):
        self._spawn()

    def _spawn(self):
        ctx = mp.get_context("spawn")
        processes = [
            Process(
                target=self.new_process,
                ctx=ctx,
                args=(i, self.lightning_module.trainer, self.mp_queue),
            )
            for i in range(self.mp_spawn_kwargs["nprocs"])
        ]
        for p in processes:
            p.start()
        while True:
            should_exit = False
            for p in processes:
                try:
                    p.watch()
                except Exception:
                    traceback.print_exc()
                    should_exit = True
            if should_exit:
                for p in processes:
                    p.terminate()
                    p.join()
                raise RuntimeError("One or more exceptions raised in sub-processes.")
            elif not all([p.is_alive() for p in processes]):
                break
            sleep(0.1)
        for p in processes:
            p.join()

    def training_step(self, *args, **kwargs):
        return self.lightning_module.training_step(*args, **kwargs)

    def validation_step(self, *args, **kwargs):
        return self.lightning_module.validation_step(*args, **kwargs)

    def test_step(self, *args, **kwargs):
        return self.lightning_module.test_step(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.lightning_module.predict(*args, **kwargs)

    def post_training_step(self):
        pass


# monkey patching, since several initialization arguments of these two plugins
# are created by Trainer->Accelerator connectors->..., and cannot be retrieved
# before the trainer is initialized.
pl.trainer.connectors.accelerator_connector.DDPPlugin = DDPPlugin
pl.trainer.connectors.accelerator_connector.DDPSpawnPlugin = DDPSpawnPlugin



from multiprocessing import Pipe
from multiprocessing.process import BaseProcess
from multiprocessing.context import _default_context
import traceback


class ProcessException(Exception):
    pass


class Process(BaseProcess):
    def __init__(
        self,
        group=None,
        target=None,
        name=None,
        args=(),
        kwargs={},
        cleaner=None,
        ctx=_default_context,
        *,
        daemon=None,
    ):
        self._exc_pipe = Pipe()
        super().__init__(
            group=group,
            target=target,
            name=name,
            args=args,
            kwargs=kwargs,
            daemon=daemon,
        )
        self._cleaner = cleaner
        self._ctx = ctx
        self._start_method = ctx.Process._start_method

    @staticmethod
    def _Popen(process_obj):
        assert isinstance(process_obj, Process)
        return process_obj._ctx.Process._Popen(process_obj)

    @property
    def exception(self):
        if not self._exc_pipe[0].poll(timeout=1e-4):
            return None
        exc = ProcessException(self._exc_pipe[0].recv())
        return exc

    def watch(self):
        if self._exc_pipe[0].poll(timeout=1e-4):
            self.join()
            raise self.exception

    @staticmethod
    def format_exceptions(exceptions):
        all_tb = ""
        for exc, i in zip(exceptions, range(len(exceptions))):
            tb = exc.__traceback__
            tb = traceback.format_exception(type(exc), exc, tb)
            tb = "".join(tb)
            all_tb = all_tb + f"\nException {i}:\n{tb}"
        return all_tb

    def run(self):
        exc = []
        try:
            super().run()
        except BaseException as e:
            exc.append(e)
        finally:
            if self._cleaner is not None:
                try:
                    self._cleaner()
                except BaseException as e:
                    exc.append(e)
            if exc:
                self._exc_pipe[1].send(self.format_exceptions(exc))



from .world import (
    World,
    CollectiveGroup,
    RpcGroup,
    get_world,
    get_cur_rank,
    get_cur_name,
    is_world_initialized,
    debug_with_process,
)

from . import world

__all__ = [
    "World",
    "CollectiveGroup",
    "RpcGroup",
    "get_world",
    "get_cur_rank",
    "get_cur_name",
    "is_world_initialized",
    "debug_with_process",
]



from machin.utils.logging import default_logger
from machin.model.nets.base import static_module_wrapper
import inspect
import torch
import torch.nn as nn


def soft_update(target_net: nn.Module, source_net: nn.Module, update_rate) -> None:
    """
    Soft update target network's parameters.

    Args:
        target_net: Target network to be updated.
        source_net: Source network providing new parameters.
        update_rate: Update rate.

    Returns:
        None
    """
    with torch.no_grad():
        for target_param, param in zip(
            target_net.parameters(), source_net.parameters()
        ):
            target_param.data.copy_(
                target_param.data * (1.0 - update_rate)
                + param.data.to(target_param.device) * update_rate
            )


def hard_update(target_net: nn.Module, source_net: nn.Module) -> None:
    """
    Hard update (directly copy) target network's parameters.

    Args:
        target_net: Target network to be updated.
        source_net: Source network providing new parameters.
    """

    for target_buffer, buffer in zip(target_net.buffers(), source_net.buffers()):
        target_buffer.data.copy_(buffer.data)
    for target_param, param in zip(target_net.parameters(), source_net.parameters()):
        target_param.data.copy_(param.data)


def determine_device(model):
    devices = set()
    for k, v in model.named_parameters():
        devices.add(str(v.device))
    return list(devices)


def safe_call(model, *named_args):
    """
    Call a model and discard unnecessary arguments. safe_call will automatically
    move tensors in named_args to the input device of the model

    Any input tensor in named_args must not be contained inside any container,
    such as list, dict, tuple, etc. Because they will be automatically moved
    to the input device of the specified model.

    Args:
        model: Model to be called, must be a wrapped nn.Module or an instance of
               NeuralNetworkModule.
        named_args: A dictionary of argument, key is argument's name, value is
                    argument's value.

    Returns:
        Whatever returned by your module. If result is not a tuple, always
        wrap results inside a tuple
    """
    org_model = None
    if isinstance(
        model, (nn.parallel.DistributedDataParallel, nn.parallel.DataParallel)
    ):
        org_model = model
        model = model.module
    if not hasattr(model, "input_device") or not hasattr(model, "output_device"):
        # try to automatically determine the input & output device of the model
        model_type = type(model)
        device = determine_device(model)
        if len(device) > 1:
            raise RuntimeError(
                f"""\
                Failed to automatically determine i/o device of your model: {model_type}
                Detected multiple devices: {device}

                You need to manually specify i/o device of your model.

                Either Wrap your model of type nn.Module with one of:
                1. static_module_wrapper from machin.model.nets.base
                2. dynamic_module_wrapper from machin.model.nets.base 
                
                Or construct your own module & model with: 
                NeuralNetworkModule from machin.model.nets.base"""
            )
        else:
            # assume that i/o devices are the same as parameter device
            # print a warning
            default_logger.warning(
                f"""\
                
                You have not specified the i/o device of your model {model_type}
                Automatically determined and set to: {device[0]}

                The framework is not responsible for any un-matching device issues 
                caused by this operation."""
            )
            model = static_module_wrapper(model, device[0], device[0])

    input_device = model.input_device
    arg_spec = inspect.getfullargspec(model.forward)
    # exclude self in arg_spec.args
    args = arg_spec.args[1:] + arg_spec.kwonlyargs
    if arg_spec.defaults is not None:
        args_with_defaults = args[-len(arg_spec.defaults) :]
    else:
        args_with_defaults = []
    required_args = (
        set(args)
        - set(args_with_defaults)
        - set(
            arg_spec.kwonlydefaults.keys()
            if arg_spec.kwonlydefaults is not None
            else []
        )
    )
    args_dict = {}

    # fill in args
    for na in named_args:
        for k, v in na.items():
            if k in args:
                if torch.is_tensor(v):
                    args_dict[k] = v.to(input_device)
                else:
                    args_dict[k] = v

    # check for necessary args
    missing = required_args - set(args_dict.keys())
    if len(missing) > 0:
        raise RuntimeError(
            f"""\
            Required arguments of the forward function of Model {type(model)} 
            is {required_args}, missing required arguments: {missing}

            Check your storage functions.
            """
        )

    if org_model is not None:
        result = org_model(**args_dict)
    else:
        result = model(**args_dict)

    if isinstance(result, tuple):
        return result
    else:
        return (result,)


def safe_return(result):
    if len(result) == 1:
        return result[0]
    else:
        return result


def safe_import(name):
    components = name.split(".")
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def get_globals_from_stack():
    frames = inspect.stack()
    global_vars = {}
    for frame in frames:
        for k, v in frame[0].f_globals.items():
            if not k.startswith("__"):
                global_vars[k] = v
    return global_vars


def assert_output_is_probs(tensor):
    if (
        tensor.dim() == 2
        and torch.all(torch.abs(torch.sum(tensor, dim=1) - 1.0) < 1e-5)
        and torch.all(tensor >= 0)
    ):
        return
    else:
        print(tensor)
        raise ValueError(
            "Input tensor is not a probability tensor, it must "
            "have 2 dimensions (0 and 1), a sum of 1.0 for each "
            "row in dimension 1, and a positive value for each "
            "element."
        )


def assert_and_get_valid_models(models):
    m = []
    global_vars = get_globals_from_stack()
    for model in models:
        if inspect.isclass(model) and issubclass(model, nn.Module):
            m.append(model)
        elif (
            isinstance(model, str)
            and model in global_vars
            and issubclass(global_vars[model], nn.Module)
        ):
            m.append(global_vars[model])
        elif isinstance(model, str) and "." in model:
            m.append(safe_import(model))
        else:
            raise ValueError(
                f"""\
                
                Invalid model: {model}, it needs to be one of:
                1. An nn.Module subclass
                2. A string name of a global defined model class in any frame 
                   of your call stack. (Not available if framework is distributed),
                3. A string name of a importable model class, eg: foo.baz.model"""
            )
    return m


def assert_and_get_valid_optimizer(optimizer):
    global_vars = get_globals_from_stack()
    if inspect.isclass(optimizer) and issubclass(optimizer, torch.optim.Optimizer):
        return optimizer
    if isinstance(optimizer, str):
        if hasattr(torch.optim, optimizer):
            return getattr(torch.optim, optimizer)
        elif optimizer in global_vars and issubclass(
            global_vars[optimizer], torch.optim.Optimizer
        ):
            return global_vars[optimizer]
        elif "." in optimizer:
            return safe_import(optimizer)
    else:
        raise ValueError(
            f"""\
            
            Invalid optimizer: {optimizer}, it needs to be one of:"
            1. An optimizer class
            2. A string name of a valid optimizer class in torch.optim
            3. A string name of a global defined optimizer class in any frame 
               of your call stack. (Not available if framework is distributed),
            4. A string name of a importable optimizer class, eg: foo.baz.optim"""
        )


def assert_and_get_valid_lr_scheduler(lr_scheduler):
    global_vars = get_globals_from_stack()
    if inspect.isclass(lr_scheduler) and issubclass(
        lr_scheduler, torch.optim.lr_scheduler._LRScheduler
    ):
        return lr_scheduler
    if isinstance(lr_scheduler, str):
        if hasattr(torch.optim.lr_scheduler, lr_scheduler):
            return getattr(torch.optim.lr_scheduler, lr_scheduler)
        elif lr_scheduler in global_vars and issubclass(
            global_vars[lr_scheduler], torch.optim.lr_scheduler._LRScheduler
        ):
            return global_vars[lr_scheduler]
        elif "." in lr_scheduler:
            return safe_import(lr_scheduler)
    else:
        raise ValueError(
            f"""\
            
            Invalid lr_scheduler: {lr_scheduler}, it needs to be one of:
            1. An lr_scheduler class
            2. A string name of a valid lr_scheduler class in torch.optim.lr_scheduler
            3. A string name of a global defined lr_scheduler class in any frame 
               of your call stack. (Not available if framework is distributed)
            4. A string name of a importable scheduler class, eg: foo.baz.lr_scheduler
            """
        )


def assert_and_get_valid_criterion(criterion):
    global_vars = get_globals_from_stack()
    if inspect.isclass(criterion) and issubclass(criterion, nn.Module):
        return criterion
    if callable(criterion):
        return criterion
    if isinstance(criterion, str):
        if hasattr(torch.nn.modules.loss, criterion):
            return getattr(torch.nn.modules.loss, criterion)
        elif criterion in global_vars and callable(global_vars[criterion]):
            return global_vars[criterion]
        elif "." in criterion:
            return safe_import(criterion)
    else:
        raise ValueError(
            f"""\
            
            Invalid lr_scheduler: {criterion}, it needs to be one of:
            1. A loss class which inherits from torch.nn.Module
            2. A callable loss function
            3. A string name of a valid loss class in torch.nn.modules.loss
            4. A string name of a global defined callable in any frame 
               of your call stack. (Not available if framework is distributed)
            5. A string name of a importable loss class, eg: foo.baz.loss"""
        )


class FakeOptimizer(torch.optim.Optimizer):
    """
    A fake optimizer which does not change model parameters.
    """

    def __init__(self, params, *_, **__):
        super().__init__(params, {})

    def step(self, *args, **kwargs):
        pass



from typing import Union, Dict, List, Tuple, Callable, Any
from copy import deepcopy
import torch as t
import torch.nn as nn
import numpy as np

from machin.frame.buffers.buffer import Transition, Buffer
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework, Config
from .utils import (
    hard_update,
    soft_update,
    safe_call,
    assert_and_get_valid_models,
    assert_and_get_valid_optimizer,
    assert_and_get_valid_criterion,
    assert_and_get_valid_lr_scheduler,
)


class DQN(TorchFramework):
    """
    DQN framework.
    """

    _is_top = ["qnet", "qnet_target"]
    _is_restorable = ["qnet_target"]

    def __init__(
        self,
        qnet: Union[NeuralNetworkModule, nn.Module],
        qnet_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict] = None,
        batch_size: int = 100,
        epsilon_decay: float = 0.9999,
        update_rate: Union[float, None] = 0.005,
        update_steps: Union[int, None] = None,
        learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        mode: str = "double",
        visualize: bool = False,
        visualize_dir: str = "",
        **__,
    ):
        """
        Note:
            DQN is only available for discrete environments.

        Note:
            Dueling DQN is a network structure rather than a framework, so
            it could be applied to all three modes.

            If ``mode = "vanilla"``, implements the simplest online DQN,
            with replay buffer.

            If ``mode = "fixed_target"``, implements DQN with a target network,
            and replay buffer. Described in `this <https://web.stanford.\
edu/class/psych209/Readings/MnihEtAlHassibis15NatureControlDeepRL.pdf>`__ essay.

            If ``mode = "double"``, implements Double DQN described in
            `this <https://arxiv.org/pdf/1509.06461.pdf>`__ essay.

        Note:
            Vanilla DQN only needs one network, so internally, ``qnet``
            is assigned to ``qnet_target``.

        Note:
            In order to implement dueling DQN, you should create two dense
            output layers.

            In your q network::

                    self.fc_adv = nn.Linear(in_features=...,
                                            out_features=num_actions)
                    self.fc_val = nn.Linear(in_features=...,
                                            out_features=1)

            Then in your ``forward()`` method, you should implement output as::

                    adv = self.fc_adv(some_input)
                    val = self.fc_val(some_input).expand(self.batch_sze,
                                                         self.num_actions)
                    return val + adv - adv.mean(1, keepdim=True)

        Note:
            Your optimizer will be called as::

                optimizer(network.parameters(), learning_rate)

            Your lr_scheduler will be called as::

                lr_scheduler(
                    optimizer,
                    *lr_scheduler_args[0],
                    **lr_scheduler_kwargs[0],
                )

            Your criterion will be called as::

                criterion(
                    target_value.view(batch_size, 1),
                    predicted_value.view(batch_size, 1)
                )

        Note:
            DQN supports two ways of updating the target network, the first
            way is polyak update (soft update), which updates the target network
            in every training step by mixing its weights with the online network
            using ``update_rate``.

            The other way is hard update, which copies weights of the online
            network after every ``update_steps`` training step.

            You can either specify ``update_rate`` or ``update_steps`` to select
            one update scheme, if both are specified, an error will be raised.

            These two different update schemes may result in different training
            stability.

        Attributes:
            epsilon: Current epsilon value, determines randomness in
                ``act_discrete_with_noise``. You can set it to any value.

        Args:
            qnet: Q network module.
            qnet_target: Target Q network module.
            optimizer: Optimizer used to optimize ``qnet``.
            criterion: Criterion used to evaluate the value loss.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            epsilon_decay: Epsilon decay rate per acting with noise step.
                ``epsilon`` attribute is multiplied with this every time
                ``act_discrete_with_noise`` is called.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            update_steps: Training step number used to update target networks.
            discount: :math:`\\gamma` used in the bellman function.
            gradient_max: Maximum gradient.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            mode: one of ``"vanilla", "fixed_target", "double"``.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.batch_size = batch_size
        self.epsilon_decay = epsilon_decay
        self.update_rate = update_rate
        self.update_steps = update_steps
        self.discount = discount
        self.grad_max = gradient_max
        self.visualize = visualize
        self.visualize_dir = visualize_dir
        self.mode = mode
        self.epsilon = 1
        self._update_counter = 0

        if mode not in {"vanilla", "fixed_target", "double"}:
            raise ValueError(f"Unknown DQN mode: {mode}")

        if update_rate is not None and update_steps is not None:
            raise ValueError(
                "You can only specify one target network update"
                " scheme, either by update_rate or update_steps,"
                " but not both."
            )

        self.qnet = qnet
        if self.mode == "vanilla":
            self.qnet_target = qnet
        else:
            self.qnet_target = qnet_target
        self.qnet_optim = optimizer(self.qnet.parameters(), lr=learning_rate)
        self.replay_buffer = (
            Buffer(replay_size, replay_device)
            if replay_buffer is None
            else replay_buffer
        )

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.qnet, self.qnet_target)

        if lr_scheduler is not None:
            if lr_scheduler_args is None:
                lr_scheduler_args = ((),)
            if lr_scheduler_kwargs is None:
                lr_scheduler_kwargs = ({},)
            self.qnet_lr_sch = lr_scheduler(
                self.qnet_optim, *lr_scheduler_args[0], **lr_scheduler_kwargs[0]
            )

        self.criterion = criterion

        super().__init__()

    @property
    def optimizers(self):
        return [self.qnet_optim]

    @optimizers.setter
    def optimizers(self, optimizers):
        self.qnet_optim = optimizers[0]

    @property
    def lr_schedulers(self):
        if hasattr(self, "qnet_lr_sch"):
            return [self.qnet_lr_sch]
        return []

    def act_discrete(self, state: Dict[str, Any], use_target: bool = False, **__):
        """
        Use Q network to produce a discrete action for
        the current state.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Action of shape ``[batch_size, 1]``.
            Any other things returned by your Q network. if they exist.
        """
        if use_target:
            result, *others = safe_call(self.qnet_target, state)
        else:
            result, *others = safe_call(self.qnet, state)

        result = t.argmax(result, dim=1).view(-1, 1)
        if len(others) == 0:
            return result
        else:
            return (result, *others)

    def act_discrete_with_noise(
        self,
        state: Dict[str, Any],
        use_target: bool = False,
        decay_epsilon: bool = True,
        **__,
    ):
        """
        Randomly selects an action from the action space according
        to a uniform distribution, with regard to the epsilon decay
        policy.

        Args:
            state: Current state.
            use_target: Whether to use the target network.
            decay_epsilon: Whether to decay the ``epsilon`` attribute.

        Returns:
            Noisy action of shape ``[batch_size, 1]``.
            Any other things returned by your Q network. if they exist.
        """
        if use_target:
            result, *others = safe_call(self.qnet_target, state)
        else:
            result, *others = safe_call(self.qnet, state)

        action_dim = result.shape[1]
        result = t.argmax(result, dim=1).view(-1, 1)

        if t.rand([1]).item() < self.epsilon:
            result = t.randint(0, action_dim, [result.shape[0], 1])

        if decay_epsilon:
            self.epsilon *= self.epsilon_decay

        if len(others) == 0:
            return result
        else:
            return (result, *others)

    def _act_discrete(self, state: Dict[str, Any], use_target: bool = False, **__):
        """
        Use Q network to produce a discrete action for
        the current state.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Action of shape ``[batch_size, 1]``
        """
        if use_target:
            result, *others = safe_call(self.qnet_target, state)
        else:
            result, *others = safe_call(self.qnet, state)
        return t.argmax(result, dim=1).view(-1, 1)

    def _criticize(self, state: Dict[str, Any], use_target: bool = False, **__):
        """
        Use Q network to evaluate current value.

        Args:
            state: Current state.
            use_target: Whether to use the target network.
        """
        if use_target:
            return safe_call(self.qnet_target, state)[0]
        else:
            return safe_call(self.qnet, state)[0]

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(
            transition,
            required_attrs=("state", "action", "reward", "next_state", "terminal"),
        )

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(
                trans,
                required_attrs=("state", "action", "reward", "next_state", "terminal"),
            )

    def update(
        self, update_value=True, update_target=True, concatenate_samples=True, **__
    ):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether update the Q network.
            update_target: Whether update targets.
            concatenate_samples: Whether concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        (
            batch_size,
            (state, action, reward, next_state, terminal, others,),
        ) = self.replay_buffer.sample_batch(
            self.batch_size,
            concatenate_samples,
            sample_method="random_unique",
            sample_attrs=["state", "action", "reward", "next_state", "terminal", "*"],
        )
        self.qnet.train()
        if self.mode == "vanilla":
            # Vanilla DQN, directly optimize q network.
            # target network is the same as the main network
            q_value = self._criticize(state)

            # gather requires long tensor, int32 is not accepted
            action_value = q_value.gather(
                dim=1,
                index=self.action_get_function(action).to(
                    device=q_value.device, dtype=t.long
                ),
            )

            target_next_q_value = (
                t.max(self._criticize(next_state), dim=1)[0].unsqueeze(1).detach()
            )
            y_i = self.reward_function(
                reward, self.discount, target_next_q_value, terminal, others
            )
            value_loss = self.criterion(action_value, y_i.type_as(action_value))

            if self.visualize:
                self.visualize_model(value_loss, "qnet", self.visualize_dir)

            if update_value:
                self.qnet.zero_grad()
                value_loss.backward()
                nn.utils.clip_grad_norm_(self.qnet.parameters(), self.grad_max)
                self.qnet_optim.step()

        elif self.mode == "fixed_target":
            # Fixed target DQN, which estimate next value by using the
            # target Q network. Similar to the idea of DDPG.
            q_value = self._criticize(state)

            # gather requires long tensor, int32 is not accepted
            action_value = q_value.gather(
                dim=1,
                index=self.action_get_function(action).to(
                    device=q_value.device, dtype=t.long
                ),
            )

            target_next_q_value = (
                t.max(self._criticize(next_state, True), dim=1)[0].unsqueeze(1).detach()
            )

            y_i = self.reward_function(
                reward, self.discount, target_next_q_value, terminal, others
            )
            value_loss = self.criterion(action_value, y_i.type_as(action_value))

            if self.visualize:
                self.visualize_model(value_loss, "qnet", self.visualize_dir)

            if update_value:
                self.qnet.zero_grad()
                self._backward(value_loss)
                nn.utils.clip_grad_norm_(self.qnet.parameters(), self.grad_max)
                self.qnet_optim.step()

            # Update target Q network
            if update_target:
                soft_update(self.qnet_target, self.qnet, self.update_rate)

        elif self.mode == "double":
            # Double DQN. DDQN also use the target network to estimate the next
            # value, but instead of selecting the maximum Q(s,a), it uses
            # the online DQN network to select an action and return Q(s,a'), to
            # reduce the over estimation.
            q_value = self._criticize(state)

            # gather requires long tensor, int32 is not accepted
            action_value = q_value.gather(
                dim=1,
                index=self.action_get_function(action).to(
                    device=q_value.device, dtype=t.long
                ),
            )

            with t.no_grad():
                target_next_q_value = self._criticize(next_state, True)
                next_action = self._act_discrete(next_state).to(
                    device=q_value.device, dtype=t.long
                )
                target_next_q_value = target_next_q_value.gather(
                    dim=1, index=next_action
                )

            y_i = self.reward_function(
                reward, self.discount, target_next_q_value, terminal, others
            )
            value_loss = self.criterion(action_value, y_i.type_as(action_value))

            if self.visualize:
                self.visualize_model(value_loss, "qnet", self.visualize_dir)

            if update_value:
                self.qnet.zero_grad()
                self._backward(value_loss)
                nn.utils.clip_grad_norm_(self.qnet.parameters(), self.grad_max)
                self.qnet_optim.step()

            # Update target Q network
            if update_target:
                if self.update_rate is not None:
                    soft_update(self.qnet_target, self.qnet, self.update_rate)
                else:
                    self._update_counter += 1
                    if self._update_counter % self.update_steps == 0:
                        hard_update(self.qnet_target, self.qnet)

        else:
            raise ValueError(f"Unknown DQN mode: {self.mode}")

        self.qnet.eval()
        # use .item() to prevent memory leakage
        return value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "qnet_lr_sch"):
            self.qnet_lr_sch.step()

    def load(self, model_dir, network_map=None, version=-1):
        # DOC INHERITED
        super().load(model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.qnet, self.qnet_target)

    @staticmethod
    def action_get_function(sampled_actions):
        """
        This function is used to get action numbers (int tensor indicating
        which discrete actions are used) from the sampled action dictionary.
        """
        return sampled_actions["action"]

    @staticmethod
    def reward_function(reward, discount, next_value, terminal, _):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * ~terminal * next_value

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "models": ["QNet", "QNet"],
            "model_args": ((), ()),
            "model_kwargs": ({}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "epsilon_decay": 0.9999,
            "update_rate": 0.005,
            "update_steps": None,
            "learning_rate": 0.001,
            "discount": 0.99,
            "gradient_max": np.inf,
            "replay_size": 500000,
            "replay_device": "cpu",
            "replay_buffer": None,
            "mode": "double",
            "visualize": False,
            "visualize_dir": "",
        }
        config = deepcopy(config)
        config["frame"] = "DQN"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = deepcopy(config["frame_config"])
        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*models, **f_config)
        return frame



from typing import Union, Dict, List, Tuple, Callable, Any
from copy import deepcopy
import torch as t
import torch.nn as nn
import numpy as np

from machin.frame.buffers.buffer import Transition, Buffer
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework, Config
from .utils import (
    hard_update,
    soft_update,
    safe_call,
    safe_return,
    assert_and_get_valid_models,
    assert_and_get_valid_optimizer,
    assert_and_get_valid_criterion,
    assert_and_get_valid_lr_scheduler,
)


class SAC(TorchFramework):
    """
    SAC framework.
    """

    _is_top = ["actor", "critic", "critic2", "critic_target", "critic2_target"]
    _is_restorable = ["actor", "critic_target", "critic2_target"]

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        critic_target: Union[NeuralNetworkModule, nn.Module],
        critic2: Union[NeuralNetworkModule, nn.Module],
        critic2_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple, Tuple, Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict, Dict, Dict] = None,
        target_entropy: float = None,
        initial_entropy_alpha: float = 1.0,
        batch_size: int = 100,
        update_rate: float = 0.005,
        update_steps: Union[int, None] = None,
        actor_learning_rate: float = 0.0005,
        critic_learning_rate: float = 0.001,
        alpha_learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        """
        See Also:
            :class:`.A2C`
            :class:`.DDPG`

        Important:
            When given a state, and an optional action, actor must
            at least return two values, similar to the actor structure
            described in :class:`.A2C`. However, when actor is asked to
            select an action based on the current state, you must make
            sure that the sampling process is **differentiable**. E.g.
            use the ``rsample`` method of torch distributions instead
            of the ``sample`` method.

            Compared to other actor-critic methods, SAC embeds the
            entropy term into its reward function directly, rather than adding
            the entropy term to actor's loss function. Therefore, we do not use
            the entropy output of your actor network.

            The SAC algorithm uses Q network as critics, so please reference
            :class:`.DDPG` for the requirements and the definition of
            ``action_trans_func``.

        Args:
            actor: Actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            critic2: The second critic network module.
            critic2_target: The second target critic network module.
            optimizer: Optimizer used to optimize ``actor``, ``critic`` and
                ``critic2``.
            criterion: Criterion used to evaluate the value loss.
            *_:
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            target_entropy: Target entropy weight :math:`\\alpha` used in
                the SAC soft value function:
                :math:`V_{soft}(s_t) = \\mathbb{E}_{q_t\\sim\\pi}[\
                                        Q_{soft}(s_t,a_t) - \
                                        \\alpha log\\pi(a_t|s_t)]`
            initial_entropy_alpha: Initial entropy weight :math:`\\alpha`
            gradient_max: Maximum gradient.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            update_steps: Training step number used to update target networks.
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
        """
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.update_steps = update_steps
        self.discount = discount
        self.visualize = visualize
        self.visualize_dir = visualize_dir
        self.entropy_alpha = t.tensor([initial_entropy_alpha], requires_grad=True)
        self.grad_max = gradient_max
        self.target_entropy = target_entropy
        self._update_counter = 0

        if update_rate is not None and update_steps is not None:
            raise ValueError(
                "You can only specify one target network update"
                " scheme, either by update_rate or update_steps,"
                " but not both."
            )

        self.actor = actor
        self.critic = critic
        self.critic_target = critic_target
        self.critic2 = critic2
        self.critic2_target = critic2_target
        self.actor_optim = optimizer(self.actor.parameters(), lr=actor_learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(), lr=critic_learning_rate)
        self.critic2_optim = optimizer(
            self.critic2.parameters(), lr=critic_learning_rate
        )
        self.alpha_optim = optimizer([self.entropy_alpha], lr=alpha_learning_rate)
        self.replay_buffer = (
            Buffer(replay_size, replay_device)
            if replay_buffer is None
            else replay_buffer
        )

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.critic, self.critic_target)
            hard_update(self.critic2, self.critic2_target)

        if lr_scheduler is not None:
            if lr_scheduler_args is None:
                lr_scheduler_args = ((), (), ())
            if lr_scheduler_kwargs is None:
                lr_scheduler_kwargs = ({}, {}, {})
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim, *lr_scheduler_args[0], **lr_scheduler_kwargs[0],
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim, *lr_scheduler_args[1], **lr_scheduler_kwargs[1]
            )
            self.critic2_lr_sch = lr_scheduler(
                self.critic2_optim, *lr_scheduler_args[1], **lr_scheduler_kwargs[1]
            )
            self.alpha_lr_sch = lr_scheduler(
                self.alpha_optim, *lr_scheduler_args[2], **lr_scheduler_kwargs[2]
            )

        self.criterion = criterion
        super().__init__()

    @property
    def optimizers(self):
        return [
            self.actor_optim,
            self.critic_optim,
            self.critic2_optim,
            self.alpha_optim,
        ]

    @optimizers.setter
    def optimizers(self, optimizers):
        (
            self.actor_optim,
            self.critic_optim,
            self.critic2_optim,
            self.alpha_optim,
        ) = optimizers

    @property
    def lr_schedulers(self):
        if (
            hasattr(self, "actor_lr_sch")
            and hasattr(self, "critic_lr_sch")
            and hasattr(self, "critic2_lr_sch")
            and hasattr(self, "alpha_lr_sch")
        ):
            return [
                self.actor_lr_sch,
                self.critic_lr_sch,
                self.critic2_lr_sch,
                self.alpha_lr_sch,
            ]
        return []

    def act(self, state: Dict[str, Any], **__):
        """
        Use actor network to produce an action for the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_return(safe_call(self.actor, state))

    def _criticize(
        self,
        state: Dict[str, Any],
        action: Dict[str, Any],
        use_target: bool = False,
        **__
    ):
        """
        Use critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Q Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(self.critic_target, state, action)[0]
        else:
            return safe_call(self.critic, state, action)[0]

    def _criticize2(
        self, state: Dict[str, Any], action: Dict[str, Any], use_target=False, **__
    ):
        """
        Use the second critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Q Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(self.critic2_target, state, action)[0]
        else:
            return safe_call(self.critic2, state, action)[0]

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(
            transition,
            required_attrs=("state", "action", "next_state", "reward", "terminal"),
        )

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(
                trans,
                required_attrs=("state", "action", "next_state", "reward", "terminal"),
            )

    def update(
        self,
        update_value=True,
        update_policy=True,
        update_target=True,
        update_entropy_alpha=True,
        concatenate_samples=True,
        **__
    ):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.
            update_target: Whether to update targets.
            update_entropy_alpha: Whether to update :math:`alpha` of entropy.
            concatenate_samples: Whether to concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        self.actor.train()
        self.critic.train()
        self.critic2.train()
        (
            batch_size,
            (state, action, reward, next_state, terminal, others,),
        ) = self.replay_buffer.sample_batch(
            self.batch_size,
            concatenate_samples,
            sample_method="random_unique",
            sample_attrs=["state", "action", "reward", "next_state", "terminal", "*"],
        )

        # Update critic network first
        with t.no_grad():
            next_action, next_action_log_prob, *_ = self.act(next_state)
            next_action = self.action_transform_function(
                next_action, next_state, others
            )
            next_value = self._criticize(next_state, next_action, True)
            next_value2 = self._criticize2(next_state, next_action, True)
            next_value = t.min(next_value, next_value2)
            next_value = next_value.view(
                batch_size, -1
            ) - self.entropy_alpha.item() * next_action_log_prob.view(batch_size, -1)
            y_i = self.reward_function(
                reward, self.discount, next_value, terminal, others
            )

        cur_value = self._criticize(state, action)
        cur_value2 = self._criticize2(state, action)
        value_loss = self.criterion(cur_value, y_i.type_as(cur_value))
        value_loss2 = self.criterion(cur_value2, y_i.type_as(cur_value))

        if self.visualize:
            self.visualize_model(value_loss, "critic", self.visualize_dir)

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            nn.utils.clip_grad_norm_(self.critic.parameters(), self.grad_max)
            self.critic_optim.step()

            self.critic2.zero_grad()
            value_loss2.backward()
            nn.utils.clip_grad_norm_(self.critic2.parameters(), self.grad_max)
            self.critic2_optim.step()

        # Update actor network
        cur_action, cur_action_log_prob, *_ = self.act(state)
        cur_action = self.action_transform_function(cur_action, state, others)
        act_value = self._criticize(state, cur_action)
        act_value2 = self._criticize2(state, cur_action)
        act_value = t.min(act_value, act_value2)

        act_policy_loss = (
            self.entropy_alpha.item() * cur_action_log_prob - act_value
        ).mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor", self.visualize_dir)

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            nn.utils.clip_grad_norm_(self.actor.parameters(), self.grad_max)
            self.actor_optim.step()

        # Update target networks
        if update_target:
            if self.update_rate is not None:
                soft_update(self.critic_target, self.critic, self.update_rate)
                soft_update(self.critic2_target, self.critic2, self.update_rate)
            else:
                self._update_counter += 1
                if self._update_counter % self.update_steps == 0:
                    hard_update(self.critic_target, self.critic)
                    hard_update(self.critic2_target, self.critic2)

        if update_entropy_alpha and self.target_entropy is not None:
            alpha_loss = -(
                t.log(self.entropy_alpha)
                * (cur_action_log_prob + self.target_entropy).cpu().detach()
            ).mean()
            self.alpha_optim.zero_grad()
            alpha_loss.backward()
            self.alpha_optim.step()
            # prevent nan
            with t.no_grad():
                self.entropy_alpha.clamp_(1e-6, 1e6)

        self.actor.eval()
        self.critic.eval()
        self.critic2.eval()
        # use .item() to prevent memory leakage
        return (-act_policy_loss.item(), (value_loss.item() + value_loss2.item()) / 2)

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()

    def load(self, model_dir, network_map=None, version=-1):
        # DOC INHERITED
        super().load(model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.critic, self.critic_target)
            hard_update(self.critic, self.critic2_target)

    @staticmethod
    def action_transform_function(raw_output_action, *_):
        return {"action": raw_output_action}

    @staticmethod
    def reward_function(reward, discount, next_value, terminal, _):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * ~terminal * next_value

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "models": ["Actor", "Critic", "Critic", "Critic", "Critic"],
            "model_args": ((), (), (), (), ()),
            "model_kwargs": ({}, {}, {}, {}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "target_entropy": None,
            "initial_entropy_alpha": 1.0,
            "batch_size": 100,
            "update_rate": 0.001,
            "update_steps": None,
            "actor_learning_rate": 0.0005,
            "critic_learning_rate": 0.001,
            "alpha_learning_rate": 0.001,
            "discount": 0.99,
            "gradient_max": np.inf,
            "replay_size": 500000,
            "replay_device": "cpu",
            "replay_buffer": None,
            "visualize": False,
            "visualize_dir": "",
        }
        config = deepcopy(config)
        config["frame"] = "SAC"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = config["frame_config"]
        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*models, **f_config)
        return frame



from typing import Union, Dict, List, Tuple, Callable, Any
from copy import deepcopy
import torch as t
import torch.nn as nn
import numpy as np

from machin.model.nets.base import NeuralNetworkModule
from machin.frame.buffers.buffer import Transition, Buffer
from .base import TorchFramework, Config
from .utils import (
    safe_call,
    assert_and_get_valid_models,
    assert_and_get_valid_optimizer,
    assert_and_get_valid_criterion,
    assert_and_get_valid_lr_scheduler,
)


class A2C(TorchFramework):
    """
    A2C framework.
    """

    _is_top = ["actor", "critic"]
    _is_restorable = ["actor", "critic"]

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple, Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict, Dict] = None,
        batch_size: int = 100,
        actor_update_times: int = 5,
        critic_update_times: int = 10,
        actor_learning_rate: float = 0.001,
        critic_learning_rate: float = 0.001,
        entropy_weight: float = None,
        value_weight: float = 0.5,
        gradient_max: float = np.inf,
        gae_lambda: float = 1.0,
        discount: float = 0.99,
        normalize_advantage: bool = True,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        """
        Important:
            when given a state, and an optional, action actor must
            at least return two values:

            **1. Action**

              For **contiguous environments**, action must be of shape
              ``[batch_size, action_dim]`` and *clamped by action space*.
              For **discrete environments**, action could be of shape
              ``[batch_size, action_dim]`` if it is a one hot vector, or
              ``[batch_size, 1]`` if it is a categorically encoded integer.

            **2. Log likelihood of action (action probability)**

              For either type of environment, log likelihood is of shape
              ``[batch_size, 1]``.

              Action probability must be differentiable, Gradient of actor
              is calculated from the gradient of action probability.

            The third entropy value is optional:

            **3. Entropy of action distribution**

              Entropy is usually calculated using dist.entropy(), its shape
              is ``[batch_size, 1]``. You must specify ``entropy_weight``
              to make it effective.

        Hint:
            For contiguous environments, action's are not directly output by
            your actor, otherwise it would be rather inconvenient to calculate
            the log probability of action. Instead, your actor network should
            output parameters for a certain distribution
            (eg: :class:`~torch.distributions.categorical.Normal`)
            and then draw action from it.

            For discrete environments,
            :class:`~torch.distributions.categorical.Categorical` is sufficient,
            since differentiable ``rsample()`` is not needed.

            This trick is also known as **reparameterization**.

        Hint:
            Actions are from samples during training in the actor critic
            family (A2C, A3C, PPO, TRPO, IMPALA).

            When your actor model is given a batch of actions and states, it
            must evaluate the states, and return the log likelihood of the
            given actions instead of re-sampling actions.

            An example of your actor in contiguous environments::

                class ActorNet(nn.Module):
                    def __init__(self):
                        super(ActorNet, self).__init__()
                        self.fc = nn.Linear(3, 100)
                        self.mu_head = nn.Linear(100, 1)
                        self.sigma_head = nn.Linear(100, 1)

                    def forward(self, state, action=None):
                        x = t.relu(self.fc(state))
                        mu = 2.0 * t.tanh(self.mu_head(x))
                        sigma = F.softplus(self.sigma_head(x))
                        dist = Normal(mu, sigma)
                        action = (action
                                  if action is not None
                                  else dist.sample())
                        action_entropy = dist.entropy()
                        action = action.clamp(-2.0, 2.0)
                        action_log_prob = dist.log_prob(action)
                        return action, action_log_prob, action_entropy

        Hint:
            Entropy weight is usually negative, to increase exploration.

            Value weight is usually 0.5. So critic network converges less
            slowly than the actor network and learns more conditions.

            Update equation is equivalent to:

            :math:`Loss= w_e * Entropy + w_v * Loss_v + w_a * Loss_a`
            :math:`Loss_a = -log\\_likelihood * advantage`
            :math:`Loss_v = criterion(target\\_bellman\\_value - V(s))`

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            actor_update_times: Times to update actor in ``update()``.
            critic_update_times: Times to update critic in ``update()``.
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            normalize_advantage: Whether to normalize the advantage function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
        """
        self.batch_size = batch_size
        self.actor_update_times = actor_update_times
        self.critic_update_times = critic_update_times
        self.discount = discount
        self.value_weight = value_weight
        self.entropy_weight = entropy_weight
        self.gradient_max = gradient_max
        self.gae_lambda = gae_lambda
        self.normalize_advantage = normalize_advantage
        self.visualize = visualize
        self.visualize_dir = visualize_dir

        self.actor = actor
        self.critic = critic
        self.actor_optim = optimizer(self.actor.parameters(), lr=actor_learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(), lr=critic_learning_rate)
        self.replay_buffer = (
            Buffer(replay_size, replay_device)
            if replay_buffer is None
            else replay_buffer
        )

        if lr_scheduler is not None:
            if lr_scheduler_args is None:
                lr_scheduler_args = ((), ())
            if lr_scheduler_kwargs is None:
                lr_scheduler_kwargs = ({}, {})
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim, *lr_scheduler_args[0], **lr_scheduler_kwargs[0],
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim, *lr_scheduler_args[1], **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion

        super().__init__()

    @property
    def optimizers(self):
        return [self.actor_optim, self.critic_optim]

    @optimizers.setter
    def optimizers(self, optimizers):
        self.actor_optim, self.critic_optim = optimizers

    @property
    def lr_schedulers(self):
        if hasattr(self, "actor_lr_sch") and hasattr(self, "critic_lr_sch"):
            return [self.actor_lr_sch, self.critic_lr_sch]
        return []

    def act(self, state: Dict[str, Any], *_, **__):
        """
        Use actor network to give a policy to the current state.

        Returns:
            Anything produced by actor.
        """
        # No need to safe_return because the number of
        # returned values is always more than one
        return safe_call(self.actor, state)

    def _eval_act(self, state: Dict[str, Any], action: Dict[str, Any], *_, **__):
        """
        Use actor network to evaluate the log-likelihood of a given
        action in the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_call(self.actor, state, action)

    def _criticize(self, state: Dict[str, Any], *_, **__):
        """
        Use critic network to evaluate current value.

        Returns:
            Value of shape ``[batch_size, 1]``
        """
        return safe_call(self.critic, state)[0]

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.

        Not suggested, since you will have to calculate "value" and "gae"
        by yourself.
        """
        self.replay_buffer.append(
            transition,
            required_attrs=(
                "state",
                "action",
                "next_state",
                "reward",
                "value",
                "gae",
                "terminal",
            ),
        )

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.

        "value" and "gae" are automatically calculated.
        """
        episode[-1]["value"] = episode[-1]["reward"]

        # calculate value for each transition
        for i in reversed(range(1, len(episode))):
            episode[i - 1]["value"] = (
                episode[i]["value"] * self.discount + episode[i - 1]["reward"]
            )

        # calculate advantage
        if self.gae_lambda == 1.0:
            for trans in episode:
                trans["gae"] = trans["value"] - self._criticize(trans["state"]).item()
        elif self.gae_lambda == 0.0:
            for trans in episode:
                trans["gae"] = (
                    trans["reward"]
                    + self.discount
                    * (1 - float(trans["terminal"]))
                    * self._criticize(trans["next_state"]).item()
                    - self._criticize(trans["state"]).item()
                )
        else:
            last_critic_value = 0
            last_gae = 0
            for trans in reversed(episode):
                critic_value = self._criticize(trans["state"]).item()
                gae_delta = (
                    trans["reward"]
                    + self.discount * last_critic_value * (1 - float(trans["terminal"]))
                    - critic_value
                )
                last_critic_value = critic_value
                last_gae = trans["gae"] = (
                    last_gae
                    * self.discount
                    * (1 - float(trans["terminal"]))
                    * self.gae_lambda
                    + gae_delta
                )

        for trans in episode:
            self.replay_buffer.append(
                trans,
                required_attrs=(
                    "state",
                    "action",
                    "next_state",
                    "reward",
                    "value",
                    "gae",
                    "terminal",
                ),
            )

    def update(
        self, update_value=True, update_policy=True, concatenate_samples=True, **__
    ):
        """
        Update network weights by sampling from buffer. Buffer
        will be cleared after update is finished.

        Args:
            update_value: Whether update the Q network.
            update_policy: Whether update the actor network.
            concatenate_samples: Whether concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        sum_act_loss = 0
        sum_value_loss = 0
        self.actor.train()
        self.critic.train()
        for _ in range(self.actor_update_times):
            # sample a batch
            batch_size, (state, action, advantage) = self.replay_buffer.sample_batch(
                self.batch_size,
                sample_method="random_unique",
                concatenate=concatenate_samples,
                sample_attrs=["state", "action", "gae"],
                additional_concat_attrs=["gae"],
            )

            # normalize advantage
            if self.normalize_advantage:
                advantage = (advantage - advantage.mean()) / (advantage.std() + 1e-6)

            if self.entropy_weight is not None:
                __, action_log_prob, new_action_entropy, *_ = self._eval_act(
                    state, action
                )
            else:
                __, action_log_prob, *_ = self._eval_act(state, action)
                new_action_entropy = None

            action_log_prob = action_log_prob.view(batch_size, 1)

            # calculate policy loss
            act_policy_loss = -(action_log_prob * advantage.type_as(action_log_prob))

            if new_action_entropy is not None:
                act_policy_loss += self.entropy_weight * new_action_entropy.mean()

            act_policy_loss = act_policy_loss.mean()
            sum_act_loss += act_policy_loss.item()

            if self.visualize:
                self.visualize_model(act_policy_loss, "actor", self.visualize_dir)

            # Update actor network
            if update_policy:
                self.actor.zero_grad()
                self._backward(act_policy_loss)
                nn.utils.clip_grad_norm_(self.actor.parameters(), self.gradient_max)
                self.actor_optim.step()

        for _ in range(self.critic_update_times):
            # sample a batch
            batch_size, (state, target_value) = self.replay_buffer.sample_batch(
                self.batch_size,
                sample_method="random_unique",
                concatenate=concatenate_samples,
                sample_attrs=["state", "value"],
                additional_concat_attrs=["value"],
            )
            # calculate value loss
            value = self._criticize(state)
            value_loss = (
                self.criterion(target_value.type_as(value), value) * self.value_weight
            )
            sum_value_loss += value_loss.item()

            if self.visualize:
                self.visualize_model(value_loss, "critic", self.visualize_dir)

            # Update critic network
            if update_value:
                self.critic.zero_grad()
                self._backward(value_loss)
                nn.utils.clip_grad_norm_(self.critic.parameters(), self.gradient_max)
                self.critic_optim.step()

        self.replay_buffer.clear()
        self.actor.eval()
        self.critic.eval()
        return (
            -sum_act_loss / self.actor_update_times,
            sum_value_loss / self.critic_update_times,
        )

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "models": ["Actor", "Critic"],
            "model_args": ((), ()),
            "model_kwargs": ({}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "actor_update_times": 5,
            "critic_update_times": 10,
            "actor_learning_rate": 0.001,
            "critic_learning_rate": 0.001,
            "entropy_weight": None,
            "value_weight": 0.5,
            "gradient_max": np.inf,
            "gae_lambda": 1.0,
            "discount": 0.99,
            "normalize_advantage": True,
            "replay_size": 500000,
            "replay_device": "cpu",
            "replay_buffer": None,
            "visualize": False,
            "visualize_dir": "",
        }
        config = deepcopy(config)
        config["frame"] = "A2C"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = deepcopy(config["frame_config"])
        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )

        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*models, **f_config)
        return frame



from machin.frame.buffers.prioritized_buffer import PrioritizedBuffer
from machin.utils.logging import default_logger

# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


class DDPGPer(DDPG):
    """
    DDPG with prioritized experience replay.

    Warning:
        Your criterion must return a tensor of shape ``[batch_size,1]``
        when given two tensors of shape ``[batch_size,1]``, since we
        need to multiply the loss with importance sampling weight
        element-wise.

        If you are using loss modules given by pytorch. It is always
        safe to use them without any modification.
    """

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        actor_target: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        critic_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple, Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict, Dict] = None,
        batch_size: int = 100,
        update_rate: float = 0.005,
        update_steps: Union[int, None] = None,
        actor_learning_rate: float = 0.0005,
        critic_learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        # DOC INHERITED
        super().__init__(
            actor,
            actor_target,
            critic,
            critic_target,
            optimizer,
            criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            update_rate=update_rate,
            update_steps=update_steps,
            actor_learning_rate=actor_learning_rate,
            critic_learning_rate=critic_learning_rate,
            discount=discount,
            gradient_max=gradient_max,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=(
                PrioritizedBuffer(replay_size, replay_device)
                if replay_buffer is None
                else replay_buffer
            ),
            visualize=visualize,
            visualize_dir=visualize_dir,
        )
        # Must have reduction attribute and value must be None
        if not hasattr(self.criterion, "reduction"):
            raise RuntimeError(
                "Criterion does not have the "
                "'reduction' property, are you using a custom "
                "criterion?"
            )
        else:
            # A loss defined in ``torch.nn.modules.loss``
            if self.criterion.reduction != "none":
                default_logger.warning(
                    "The reduction property of criterion is not 'none', "
                    "automatically corrected."
                )
                self.criterion.reduction = "none"

    def update(
        self,
        update_value=True,
        update_policy=True,
        update_target=True,
        concatenate_samples=True,
        **__
    ):
        # DOC INHERITED
        self.actor.train()
        self.critic.train()
        (
            batch_size,
            (state, action, reward, next_state, terminal, others),
            index,
            is_weight,
        ) = self.replay_buffer.sample_batch(
            self.batch_size,
            concatenate_samples,
            sample_attrs=["state", "action", "reward", "next_state", "terminal", "*"],
        )

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_function(
                self._act(next_state, True), next_state, others
            )
            next_value = self._criticize(next_state, next_action, True)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_function(
                reward, self.discount, next_value, terminal, others
            )

        # critic loss
        cur_value = self._criticize(state, action)
        value_loss = self.criterion(cur_value, y_i.type_as(cur_value))
        value_loss = value_loss * t.from_numpy(is_weight).view([batch_size, 1]).type_as(
            value_loss
        )
        value_loss = value_loss.mean()

        if self.visualize:
            self.visualize_model(value_loss, "critic", self.visualize_dir)

        if update_value:
            self.critic.zero_grad()
            self._backward(value_loss)
            nn.utils.clip_grad_norm_(self.critic.parameters(), self.gradient_max)
            self.critic_optim.step()

        # actor loss
        cur_action = self.action_transform_function(self._act(state), state, others)
        act_value = self._criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        # update priority
        abs_error = (
            t.sum(t.abs(act_value - y_i.type_as(act_value)), dim=1)
            .flatten()
            .detach()
            .cpu()
            .numpy()
        )
        self.replay_buffer.update_priority(abs_error, index)

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor", self.visualize_dir)

        if update_policy:
            self.actor.zero_grad()
            self._backward(act_policy_loss)
            nn.utils.clip_grad_norm_(self.actor.parameters(), self.gradient_max)
            self.actor_optim.step()

        # Update target networks
        if update_target:
            if self.update_rate is not None:
                soft_update(self.actor_target, self.actor, self.update_rate)
                soft_update(self.critic_target, self.critic, self.update_rate)
            else:
                self._update_counter += 1
                if self._update_counter % self.update_steps == 0:
                    hard_update(self.actor_target, self.actor)
                    hard_update(self.critic_target, self.critic)

        self.actor.eval()
        self.critic.eval()
        # use .item() to prevent memory leakage
        return -act_policy_loss.item(), value_loss.item()

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        config = DDPG.generate_config(config)
        config["frame"] = "DDPGPer"
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = config["frame_config"]
        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        criterion.reduction = "none"
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*models, **f_config)
        return frame



from machin.frame.buffers.prioritized_buffer import PrioritizedBuffer
from machin.utils.logging import default_logger

# pylint: disable=wildcard-import, unused-wildcard-import
from .dqn import *


class DQNPer(DQN):
    """
    DQN with prioritized replay. It is based on Double DQN.

    Warning:
        Your criterion must return a tensor of shape ``[batch_size,1]``
        when given two tensors of shape ``[batch_size,1]``, since we
        need to multiply the loss with importance sampling weight
        element-wise.

        If you are using loss modules given by pytorch. It is always
        safe to use them without any modification.
    """

    def __init__(
        self,
        qnet: Union[NeuralNetworkModule, nn.Module],
        qnet_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple] = None,
        lr_scheduler_kwargs: Tuple[Dict] = None,
        batch_size: int = 100,
        epsilon_decay: float = 0.9999,
        update_rate: Union[float, None] = 0.005,
        update_steps: Union[int, None] = None,
        learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        # DOC INHERITED
        super().__init__(
            qnet,
            qnet_target,
            optimizer,
            criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            epsilon_decay=epsilon_decay,
            update_rate=update_rate,
            update_steps=update_steps,
            learning_rate=learning_rate,
            discount=discount,
            gradient_max=gradient_max,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=(
                PrioritizedBuffer(replay_size, replay_device)
                if replay_buffer is None
                else replay_buffer
            ),
            mode="double",
            visualize=visualize,
            visualize_dir=visualize_dir,
        )
        # reduction must be None
        if not hasattr(self.criterion, "reduction"):
            raise RuntimeError(
                "Criterion does not have the "
                "'reduction' property, are you using a custom "
                "criterion?"
            )
        else:
            # A loss defined in ``torch.nn.modules.loss``
            if self.criterion.reduction != "none":
                default_logger.warning(
                    "The reduction property of criterion is not 'none', "
                    "automatically corrected."
                )
                self.criterion.reduction = "none"

    def update(
        self, update_value=True, update_target=True, concatenate_samples=True, **__
    ):
        # DOC INHERITED
        self.qnet.train()
        (
            batch_size,
            (state, action, reward, next_state, terminal, others),
            index,
            is_weight,
        ) = self.replay_buffer.sample_batch(
            self.batch_size,
            concatenate_samples,
            sample_method="random_unique",
            sample_attrs=["state", "action", "reward", "next_state", "terminal", "*"],
        )

        with t.no_grad():
            next_q_value = self._criticize(next_state)
            target_next_q_value = self._criticize(next_state, True)
            target_next_q_value = target_next_q_value.gather(
                dim=1, index=t.max(next_q_value, dim=1)[1].unsqueeze(1)
            )

        q_value = self._criticize(state)

        # gather requires long tensor, int32 is not accepted
        action_value = q_value.gather(
            dim=1,
            index=self.action_get_function(action).to(
                device=q_value.device, dtype=t.long
            ),
        )

        # Generate value reference :math: `y_i`.
        y_i = self.reward_function(
            reward, self.discount, target_next_q_value, terminal, others
        )
        value_loss = self.criterion(action_value, y_i.to(action_value.device))
        value_loss = value_loss * t.from_numpy(is_weight).view([batch_size, 1]).to(
            value_loss.device
        )
        value_loss = value_loss.mean()

        abs_error = (
            t.sum(t.abs(action_value - y_i.to(action_value.device)), dim=1)
            .flatten()
            .detach()
            .cpu()
            .numpy()
        )
        self.replay_buffer.update_priority(abs_error, index)

        if self.visualize:
            self.visualize_model(value_loss, "qnet", self.visualize_dir)

        if update_value:
            self.qnet.zero_grad()
            self._backward(value_loss)
            nn.utils.clip_grad_norm_(self.qnet.parameters(), self.grad_max)
            self.qnet_optim.step()

        # Update target Q network
        if update_target:
            if self.update_rate is not None:
                soft_update(self.qnet_target, self.qnet, self.update_rate)
            else:
                self._update_counter += 1
                if self._update_counter % self.update_steps == 0:
                    hard_update(self.qnet_target, self.qnet)

        self.qnet.eval()
        # use .item() to prevent memory leakage
        return value_loss.item()

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        config = DQN.generate_config(config)
        config["frame"] = "DQNPer"
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = deepcopy(config["frame_config"])
        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        criterion.reduction = "none"
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*models, **f_config)
        return frame



from typing import Union, Dict, List, Any
from threading import RLock
from collections import OrderedDict
from ..transition import Transition
from .prioritized_buffer import PrioritizedBuffer
from machin.parallel.distributed import RpcGroup
import numpy as np
import torch as t


class DistributedPrioritizedBuffer(PrioritizedBuffer):
    def __init__(self, buffer_name: str, group: RpcGroup, buffer_size: int, *_, **__):
        """
        Create a distributed prioritized replay buffer instance.

        To avoid issues caused by tensor device difference, all transition
        objects are stored in device "cpu".

        Distributed prioritized replay buffer constitutes of many local buffers
        held per process, since it is very inefficient to maintain a weight
        tree across processes, each process holds a weight tree of records in
        its local buffer and a local buffer (same as ``DistributedBuffer``).

        The sampling process(es) will first use rpc to acquire the wr_lock,
        signalling "stop" to appending performed by actor processes,
        then perform a sum of all local weight trees, and finally perform
        sampling, after sampling and updating the importance weight,
        the lock will be released.


        During sampling, the tensors in "state", "action" and "next_state"
        dictionaries, along with "reward", will be concatenated in dimension 0.
        any other custom keys specified in ``**kwargs`` will not be
        concatenated.

        .. seealso:: :class:`PrioritizedBuffer`

        Note:
            :class:`DistributedPrioritizedBuffer` is not split into an
            accessor and an implementation, because we would like to operate
            on the buffer directly, when calling "size()" or "append()", to
            increase efficiency (since rpc layer is bypassed).

        Args:
            buffer_size: Maximum local buffer size.
            group: Process group which holds this buffer.
        """
        super().__init__(buffer_size, "cpu")
        self.buffer_name = buffer_name
        self.buffer_version_table = np.zeros([buffer_size], dtype=np.uint64)
        self.group = group

        assert group.is_member()

        # register services, so that we may access other buffers
        _name = "/" + group.get_cur_name()
        self.group.register(buffer_name + _name + "/_size_service", self._size_service)
        self.group.register(
            buffer_name + _name + "/_clear_service", self._clear_service
        )
        self.group.register(
            buffer_name + _name + "/_weight_sum_service", self._weight_sum_service
        )
        self.group.register(
            buffer_name + _name + "/_update_priority_service",
            self._update_priority_service,
        )
        self.group.register(
            buffer_name + _name + "/_sample_service", self._sample_service
        )
        self.wr_lock = RLock()

    def append(
        self,
        transition: Union[Transition, Dict],
        priority: Union[float, None] = None,
        required_attrs=("state", "action", "next_state", "reward", "terminal"),
    ):
        # DOC INHERITED
        with self.wr_lock:
            position = super(PrioritizedBuffer, self).append(transition, required_attrs)
            if priority is None:
                # the initialization method used in the original essay
                priority = self.wt_tree.get_leaf_max()
            self.wt_tree.update_leaf(self._normalize_priority(priority), position)
            # increase the version counter to mark it as tainted
            # later priority update will ignore this position
            self.buffer_version_table[position] += 1

    def size(self):
        """
        Returns:
            Length of current local buffer.
        """
        with self.wr_lock:
            return super().size()

    def all_size(self):
        """
        Returns:
            Total length of all buffers.
        """
        future = []
        count = 0
        for m in self.group.get_group_members():
            future.append(
                self.group.registered_async(
                    self.buffer_name + "/" + m + "/_size_service"
                )
            )
        for fut in future:
            count += fut.wait()
        return count

    def clear(self):
        """
        Remove all entries from current local buffer.
        """
        with self.wr_lock:
            super().clear()
            # also clear the version table
            self.buffer_version_table.fill(0)

    def all_clear(self):
        """
        Remove all entries from all local buffers.
        """
        future = [
            self.group.registered_async(self.buffer_name + "/" + m + "/_clear_service")
            for m in self.group.get_group_members()
        ]
        for fut in future:
            fut.wait()

    def update_priority(self, priorities: np.ndarray, indexes: OrderedDict):
        # DOC INHERITED
        # update priority on all local buffers
        future = []
        offset = 0

        # indexes is an OrderedDict, key is sampled process name,
        # value is a tuple of an index np.ndarray and a version np.ndarray
        for m, sub in indexes.items():
            length = len(sub[0])
            future.append(
                self.group.registered_async(
                    self.buffer_name + "/" + m + "/_update_priority_service",
                    args=(priorities[offset : offset + length], sub[0], sub[1]),
                )
            )
            offset += length
        for fut in future:
            fut.wait()

    def sample_batch(
        self,
        batch_size: int,
        concatenate: bool = True,
        device: Union[str, t.device] = None,
        sample_attrs: List[str] = None,
        additional_concat_attrs: List[str] = None,
        *_,
        **__
    ) -> Any:
        if batch_size <= 0:
            return 0, None, None, None

        # calculate all weight sum
        future = [
            self.group.registered_async(
                self.buffer_name + "/" + m + "/_weight_sum_service"
            )
            for m in self.group.get_group_members()
        ]
        weights = [fut.wait() for fut in future]
        all_weight_sum = sum(weights) + 1e-6  # prevent all zero

        # determine the sampling size of local buffers, based on:
        # local_weight_sum / all_weight_sum
        ssize = np.ceil(np.array(weights) * batch_size / all_weight_sum)
        ssize = [int(ss) for ss in ssize]
        # collect samples and their priority
        future = [
            (
                m,
                self.group.registered_async(
                    self.buffer_name + "/" + m + "/_sample_service",
                    args=(ss, all_weight_sum),
                ),
            )
            for m, ss in zip(self.group.get_group_members(), ssize)
        ]

        all_batch_len = 0
        all_batch = []
        all_index = OrderedDict()
        all_is_weight = []
        for m, fut in future:
            batch_len, batch, index, version, is_weight = fut.wait()
            if batch_len == 0:
                continue
            all_batch_len += batch_len
            all_batch += batch
            all_is_weight.append(is_weight)
            # store them together to make API compatible with PrioritizedBuffer
            all_index[m] = (index, version)
        if all_batch_len == 0:
            return 0, None, None, None
        all_batch = PrioritizedBuffer.post_process_batch(
            all_batch, device, concatenate, sample_attrs, additional_concat_attrs
        )
        all_is_weight = np.concatenate(all_is_weight, axis=0)
        return all_batch_len, all_batch, all_index, all_is_weight

    def _size_service(self):  # pragma: no cover
        with self.wr_lock:
            return super().size()

    def _clear_service(self):  # pragma: no cover
        with self.wr_lock:
            super().clear()
            # also clear the version table
            self.buffer_version_table.fill(0)

    def _weight_sum_service(self):  # pragma: no cover
        with self.wr_lock:
            return self.wt_tree.get_weight_sum()

    def _update_priority_service(
        self, priorities, indexes, versions
    ):  # pragma: no cover
        with self.wr_lock:
            # compare original entry versions to the current version table
            is_same = self.buffer_version_table[indexes] == versions
            # select unchanged entries
            priorities = priorities[is_same]
            indexes = indexes[is_same]
            super().update_priority(priorities, indexes)

    def _sample_service(self, batch_size, all_weight_sum):  # pragma: no cover
        # the local batch size
        with self.wr_lock:
            if batch_size <= 0 or len(self.buffer) == 0:
                return 0, None, None, None, None

            wt_tree = self.wt_tree

            segment_length = wt_tree.get_weight_sum() / batch_size
            rand_priority = np.random.uniform(size=batch_size) * segment_length
            rand_priority += np.arange(batch_size, dtype=np.float) * segment_length
            rand_priority = np.clip(
                rand_priority, 0, max(wt_tree.get_weight_sum() - 1e-6, 0)
            )
            index = wt_tree.find_leaf_index(rand_priority)
            version = self.buffer_version_table[index]

            batch = [self.buffer[idx] for idx in index]
            priority = wt_tree.get_leaf_weight(index)

            # calculate importance sampling weight
            sample_probability = priority / all_weight_sum
            is_weight = np.power(len(self.buffer) * sample_probability, -self.curr_beta)
            is_weight /= is_weight.max()
            self.curr_beta = np.min(
                [1.0, self.curr_beta + self.beta_increment_per_sampling]
            )
            return len(batch), batch, index, version, is_weight



from typing import Callable, Tuple, Any, Dict
import torch as t
import torch.nn as nn

from machin.utils.logging import default_logger as logger
from machin.utils.helper_classes import Switch
from .generator import NormalNoiseGen


class AdaptiveParamNoise:
    def __init__(
        self,
        initial_stddev: float = 0.1,
        desired_action_stddev: float = 0.1,
        adoption_coefficient: float = 1.01,
    ):
        """
        Implements the adaptive parameter space method in
        `<<Parameter space noise for exploration>> \
<https://arxiv.org/pdf/1706.01905.pdf>`_.

        Hint:
            Let :math:`\\theta` be the standard deviation of noise,
            and :math:`\\alpha` be the adpotion coefficient, then:

            :math:`\\theta_{n+1} = \\left \\{ \
                \\begin{array}{ll} \
                    \\alpha \\theta_k \
                        & if\\ d(\\pi,\\tilde{\\pi})\\leq\\delta, \\\\ \
                    \\frac{1}{\\alpha} \\theta_k & otherwise, \
                \\end{array} \
            \\right. \\ `

            Noise is directly applied to network parameters.

        Args:
            initial_stddev: Initial noise standard deviation.
            desired_action_stddev: Desired standard deviation for
            adoption_coefficient: Adoption coefficient.
        """

        self.initial_stddev = initial_stddev
        self.desired_action_stddev = desired_action_stddev
        self.adoption_coefficient = adoption_coefficient

        self.current_stddev = initial_stddev

    def adapt(self, distance: float):
        """
        Update noise standard deviation according to distance.

        Args:
            distance: Current distance between the noisy action and clean
                action.
        """
        if distance > self.desired_action_stddev:
            # Decrease stddev.
            self.current_stddev /= self.adoption_coefficient
        else:
            # Increase stddev.
            self.current_stddev *= self.adoption_coefficient

    def get_dev(self) -> float:
        """
        Returns:
            Current noise standard deviation.
        """
        return self.current_stddev

    def __repr__(self):
        fmt = "AdaptiveParamNoise(i_std={}, da_std={}, adpt_coeff={})"
        return fmt.format(
            self.initial_stddev, self.desired_action_stddev, self.adoption_coefficient
        )


def _add_perturb_hook(
    module, perturb_switch, reset_switch, perturb_gen, debug_backward
):
    org_params = {}
    noisy_params = {}

    def perturb_pre_hook(*_):
        with t.no_grad():
            if perturb_switch.get():
                if noisy_params and not reset_switch.get():
                    # Use generated noisy parameters.
                    for p_name, p_value in module.named_parameters():
                        if t.is_tensor(p_value):
                            p_value.set_(noisy_params[p_name])
                else:
                    # Generate noisy parameters if they don't exist.
                    org_params.clear()
                    noisy_params.clear()
                    for p_name, p_value in module.named_parameters():
                        if t.is_tensor(p_value):
                            org_params[p_name] = p_value.clone()
                            p_value += perturb_gen(
                                p_value.shape, p_value.device
                            ).detach()
                            noisy_params[p_name] = p_value.clone()

            elif not perturb_switch.get():
                # Use original parameters
                if org_params:
                    for p_name, p_value in module.named_parameters():
                        if t.is_tensor(p_value):
                            p_value.set_(org_params[p_name])

    pre_hook_handle = module.register_forward_pre_hook(perturb_pre_hook)

    post_hook_handles = []
    for param_name, param_value in module.named_parameters():

        def perturb_post_hook(*_):  # pragma: no cover
            # pytest will not detect execution by autograd engine
            # Called before backward update, swap noisy parameters out,
            # so gradients are applied to original parameters.
            if debug_backward:
                print(f"Backward swapped for {param_name}!")
            with t.no_grad():
                if org_params and t.is_tensor(param_value):
                    param_value.set_(org_params[param_name])

        post_hook_handle = param_value.register_hook(perturb_post_hook)
        post_hook_handles.append(post_hook_handle)

    return pre_hook_handle, post_hook_handles


# noinspection PyTypeChecker
def perturb_model(
    model: nn.Module,
    perturb_switch: Switch,
    reset_switch: Switch,
    distance_func: Callable = lambda x, y: t.dist(x, y, 2).mean().item(),
    desired_action_stddev: float = 0.5,
    noise_generator: Any = NormalNoiseGen,
    noise_generator_args: Tuple = (),
    noise_generator_kwargs: Dict = None,
    noise_generate_function: Callable = None,
    debug_backward=False,
):
    """
    Give model's parameters a little perturbation. Implements
    `<<Parameter space noise for exploration>> \
<https://arxiv.org/pdf/1706.01905.pdf>`_.

    Note:
        Only parameters of type ``t.Tensor`` and gettable from
        ``model.named_parameters()`` will be perturbed.

        Original parameters will be automatically swapped in during the
        backward pass, and you can safely call optimizers afterwards.

    Hint:
        1. ``noise_generator`` must accept (shape, \\*args) in its ``__init__``
        function, where shape is the required shape. it also needs to have
        ``__call__(device=None)`` which produce a noise tensor on the specified
        device when invoked.

        2. ``noise_generate_function`` must accept (shape, device, std:float)
        and return a noise tensor on the specified device.

    Example:
        In order to use this function to perturb your model, you need to::

            from machin.utils.helper_classes import Switch
            from machin.frame.noise.param_space_noise import perturb_model
            from machin.utils.visualize import visualize_graph
            import torch as t

            dims = 5

            t.manual_seed(0)
            model = t.nn.Linear(dims, dims)
            optim = t.optim.Adam(model.parameters(), 1e-3)
            p_switch, r_switch = Switch(), Switch()
            cancel = perturb_model(model, p_switch, r_switch)

            # you should keep this switch on if you do one training step after
            # every sampling step. otherwise you may turn it off in one episode
            # and turn it on in the next to speed up training.
            r_switch.on()

            # turn off/on the perturbation switch to see the difference
            p_switch.on()

            # do some sampling
            action = model(t.ones([dims]))

            # in order to let parameter noise adapt to generate noisy actions
            # within ``desired_action_stddev``, you must periodically
            # use the original model to generate some actions:
            p_switch.off()
            action = model(t.ones([dims]))

            # visualize will not show any leaf noise tensors
            # because they are created in t.no_grad() context
            # and added in-place.
            visualize_graph(action, exit_after_vis=False)

            # do some training
            loss = (action - t.ones([dims])).sum()
            loss.backward()
            optim.step()
            print(model.weight)

            # clear hooks
            cancel()

    Args:
        model: Neural network model.
        perturb_switch: The switch used to enable perturbation. If switch is
            set to ``False`` (off), then during the forward process, original
            parameters are used.
        reset_switch: The switch used to reset perturbation noise. If switch is
            set to ``True`` (on), and ``perturb_switch`` is also on, then during
            every forward process, a new set of noise is applied to each param.
            If only ``perturb_switch`` is on, then the same set of noisy
            parameters is used in the forward process and they **will not be
            updated**.
        distance_func: Distance function, accepts two tensors produced by
            ``model`` (one is noisy), return the distance as float. Used
            to compare the distance between actions generated by
            noisy parameters and original parameters.
        desired_action_stddev: Desired action standard deviation.
        noise_generator: Noise generator class.
        noise_generator_args: Additional args other than shape of the noise
            generator.
        noise_generator_kwargs: Additional kwargs other than shape of the noise
            generator.
        noise_generate_function: Noise generation function, mutually exclusive
            with ``noise_generator`` and ``noise_generator_args``.
        debug_backward: Print a message if the backward hook is correctly
            executed.

    Returns:
        1. A reset function with no arguments, will swap in original paramters.
        2. A deregister function with no arguments, will deregister all hooks
            applied on your model.
    """
    tmp_action = {}
    hook_handles = []

    param_noise_spec = AdaptiveParamNoise(desired_action_stddev=desired_action_stddev)

    def param_noise_gen(shape, device):
        nonlocal noise_generator_args, noise_generator_kwargs
        if noise_generator_kwargs is None:
            noise_generator_kwargs = {}
        gen = noise_generator(shape, *noise_generator_args, **noise_generator_kwargs)
        return gen(device) * param_noise_spec.get_dev()

    def param_noise_custom_gen_wrapper(shape, device):
        std_dev = param_noise_spec.get_dev()
        return noise_generate_function(shape, device, std_dev)

    if noise_generate_function is not None:
        param_noise_gen = param_noise_custom_gen_wrapper

    def perturb_adjust_hook(_model, _input, output):
        if perturb_switch.get():
            tmp_action["with_noise"] = output.clone()
        else:
            tmp_action["without_noise"] = output.clone()
        if "with_noise" in tmp_action and "without_noise" in tmp_action:
            # Compute distance between two actions generated by
            # noisy parameters and original parameters.
            with t.no_grad():
                dist = distance_func(
                    tmp_action["with_noise"], tmp_action["without_noise"]
                )
                tmp_action.clear()
                param_noise_spec.adapt(dist)
                logger.info(f"Current output distance: {dist}")
                logger.info(f"Current param noise stddev: {param_noise_spec.get_dev()}")

    # Boise generation happens in pre-forward and noise adjust happens
    # in post-forward
    hook_handles.append(model.register_forward_hook(perturb_adjust_hook))

    pre, post = _add_perturb_hook(
        model, perturb_switch, reset_switch, param_noise_gen, debug_backward
    )
    hook_handles.append(pre)
    hook_handles += post

    def cancel():
        for hh in hook_handles:
            hh.remove()

    return cancel



from . import ordered_server
from . import param_server
from .ordered_server import (
    OrderedServerBase,
    OrderedServerSimple,
    OrderedServerSimpleImpl,
)
from .param_server import (
    PushPullGradServer,
    PushPullGradServerImpl,
    PushPullModelServer,
    PushPullModelServerImpl,
)

__all__ = [
    "OrderedServerBase",
    "OrderedServerSimple",
    "OrderedServerSimpleImpl",
    "PushPullGradServer",
    "PushPullGradServerImpl",
    "PushPullModelServer",
    "PushPullModelServerImpl",
    "ordered_server",
    "param_server",
]



import warnings
from .base import TorchFramework

from .dqn import DQN
from .dqn_per import DQNPer
from .rainbow import RAINBOW

from .ddpg import DDPG
from .hddpg import HDDPG
from .td3 import TD3
from .ddpg_per import DDPGPer

from .a2c import A2C
from .a3c import A3C
from .ppo import PPO
from .sac import SAC

from .maddpg import MADDPG

try:
    from .apex import DQNApex, DDPGApex
    from .impala import IMPALA
    from .ars import ARS
except ImportError as _:
    warnings.warn(
        "Failed to import algorithms relying on torch.distributed." " Set them to None."
    )
    DQNApex = None
    DDPGApex = None
    IMPALA = None
    ARS = None


__all__ = [
    "TorchFramework",
    "DQN",
    "DQNPer",
    "RAINBOW",
    "DDPG",
    "HDDPG",
    "TD3",
    "DDPGPer",
    "A2C",
    "A3C",
    "PPO",
    "SAC",
    "DQNApex",
    "DDPGApex",
    "IMPALA",
    "ARS",
    "MADDPG",
]
